# coding: utf-8
# pylint: disable=c-extension-no-member,no-name-in-module,import-error

import logging
import os
import sys
import threading
import traceback

BASE_DIR = os.path.dirname(__file__)

try:
    from tionix_vdi_client.version import __version__, __os_bitness__
except ImportError:
    __version__ = '-'
    __os_bitness__ = ''

MODULE_VERSION = __version__
WIN_OS_BITNESS = __os_bitness__

TRACEBACK_ENABLED = 'TNX_DEBUG' in os.environ


def handle_exception(exc_type, exc_value, exc_traceback):
    if issubclass(exc_type, KeyboardInterrupt):
        sys.__excepthook__(exc_type, exc_value, exc_traceback)
        return

    logger = logging.getLogger(__name__)
    stack = traceback.extract_tb(exc_traceback)
    filepath, lineno = stack[0][:2]
    for frame in stack[::-1]:
        filepath, lineno = frame[:2]
        if 'tionix_vdi_client' in filepath:
            break
    filename = os.path.split(filepath)[-1]
    logger.error("Error in %s:%s: %s:%s",
                 filename, lineno, exc_type.__name__, exc_value,
                 exc_info=TRACEBACK_ENABLED)

    if threading.current_thread().__class__.__name__ == '_MainThread':
        try:
            from PyQt5.QtWidgets import QApplication
            from tionix_vdi_client.gui.messagebox import detailed_error
            from tionix_vdi_client.log import get_log_msg
            from tionix_vdi_client.exceptions import ErrorCodes
            from tionix_vdi_client.i18n import ugettext as _

            if QApplication.instance():
                msg = ErrorCodes.MESSAGES.get(ErrorCodes.UNEXPECTED_ERROR)
                detailed_error(None, title=_('Error'), text=msg,
                               detailed=get_log_msg())
        except Exception as err:
            logger.error('Failed to show error messagebox: %s:%s',
                         err.__class__.__name__, err)


sys.excepthook = handle_exception


CONNECTION_TYPE = 'client'
CLIENT = 'TIONIX.VDIclient {0}'.format(MODULE_VERSION)
EX_OK = 0
EX_SOFTWARE = 1


class Localization(object):
    EN = 'en'
    RU = 'ru'

    RU_CODE = 'ru_RU'
    EN_US_CODE = 'en_US'

    locales_dict = {
        RU: RU_CODE,
        EN: EN_US_CODE,
    }

    locales_translate = {
        RU: 'Русский',
        EN: 'English',
    }

    # pylint: disable=inconsistent-return-statements
    @classmethod
    def get_code_by_locale(cls, locale):
        for key, value in list(cls.locales_translate.items()):
            if value == locale:
                return key

    # pylint: disable=inconsistent-return-statements
    @classmethod
    def get_lang_by_code(cls, code):
        for key, value in list(cls.locales_dict.items()):
            if value == code:
                return key
        return cls.EN
