# coding: utf-8
import getpass
import logging
import sys
import time

from tionix_vdi_client import (
    exceptions, TRACEBACK_ENABLED, EX_OK, EX_SOFTWARE, client)
from tionix_vdi_client.exceptions import ErrorCodes
from tionix_vdi_client.helpers import RequestStatus
from tionix_vdi_client.i18n import ugettext as _
from tionix_vdi_client.rdp import rdp_connect
from tionix_vdi_client.settings import SESSION, CONF

logger = logging.getLogger(__name__)


# pylint: disable=too-many-return-statements
def main():
    conn_credentials = {
        'user': SESSION.user,
        'password': SESSION.password,
        'domain_name': CONF.domain_name,
    }
    if CONF.project:
        conn_credentials['project'] = SESSION.project
    if CONF.ikecfg:
        conn_credentials['ip_address'] = CONF.ikecfg

    if CONF.store_session:
        SESSION.save()

    try:
        vdi_client = client.VdiClient(**conn_credentials)
    except exceptions.MissingParameterException as err:
        logger.error(str(err))
        return EX_SOFTWARE

    try:
        request_id = vdi_client.request_vm()
        while True:
            vm = vdi_client.get_vm(request_id)

            if vm.status == RequestStatus.COMPLETED:
                vm_ip = vm.info['ip']
                break
            else:
                time.sleep(CONF.get_vm_timeout)

        rdp_connect(
            user=vdi_client.auth_params['user'],
            password=vdi_client.auth_params['password'],
            domain_name=vdi_client.auth_params.get('user_domain_name', ''),
            vm_ip=vm_ip,
            rdp_token=vm.rdp_token or ''
        )

    except exceptions.VDIServerResponseError as err:
        if err.code not in (exceptions.ErrorCodes.PASSWORD_EXPIRED,
                            exceptions.ErrorCodes.PASSWORD_EXPIRE_WARNING,
                            exceptions.ErrorCodes.PASSWORD_ATTEMPTS_WARNING):
            message = (
                ErrorCodes.MESSAGES.get(err.code) or
                ErrorCodes.MESSAGES.get(ErrorCodes.UNEXPECTED_ERROR))
            logger.error(message)
            return EX_SOFTWARE
        msg = _("The password needs to be changed.")
        logger.info(msg)
        new_password = getpass.getpass(prompt="New password: ")
        confirm_new_password = getpass.getpass(prompt="Confirm new password: ")

        if not new_password:
            logger.error(_('Password could not be empty.'))
            return EX_SOFTWARE

        if new_password != confirm_new_password:
            logger.error(_('Passwords do not match. '
                           'Make sure they are the same.'))
            return EX_SOFTWARE

        try:
            vdi_client.change_password(new_password)
        except exceptions.VDIException as err:
            logger.error(err, exc_info=TRACEBACK_ENABLED)
            return EX_SOFTWARE
        else:
            msg = _("Password successfully changed.")
            logger.info(msg)
    except exceptions.ApplicationLogicException as err:
        return err.exit_code
    except Exception as err:
        logger.error(err, exc_info=TRACEBACK_ENABLED)
        return EX_SOFTWARE

    return EX_OK


if __name__ == "__main__":
    sys.exit(main())
