# coding: utf-8
# pylint: disable=c-extension-no-member,no-name-in-module,import-error


import atexit
import logging
import sys

from PyQt5 import QtCore
from PyQt5.QtWidgets import QMessageBox

from tionix_vdi_client import EX_SOFTWARE, EX_OK, Localization
from tionix_vdi_client.gui.application import QSingleApplication
from tionix_vdi_client.gui.main_window import MainWindow
from tionix_vdi_client.gui.messagebox import detailed_error
from tionix_vdi_client.gui.password_window import PasswordWindow
from tionix_vdi_client.gui.pin_window import PinWindow
from tionix_vdi_client.gui.request_window import InstanceRequestWindow
from tionix_vdi_client.i18n import ugettext as _
from tionix_vdi_client.log import get_log_msg
from tionix_vdi_client.settings import CONF, SESSION
from tionix_vdi_client.smart_card import api as smartcard_api
from tionix_vdi_client.smart_card import exceptions as smartcard_exc
from tionix_vdi_client.smart_card.exceptions import IncorrectPin
from tionix_vdi_client.identity import request_token
from tionix_vdi_client.exceptions import ApplicationConnectionError

logger = logging.getLogger(__name__)


# pylint: disable=too-many-return-statements,too-many-statements
# noinspection PyTypeChecker
def main():
    app = QSingleApplication([])
    if CONF.language == Localization.RU:
        qt_translator = QtCore.QTranslator(app)
        qt_translator.load(':/i18n/qt_ru.qm')
        app.installTranslator(qt_translator)

        vdi_translator = QtCore.QTranslator()
        vdi_translator.load(":/i18n/vdi_ru.qm")
        app.installTranslator(vdi_translator)

    if CONF.use_cert:
        if CONF.use_smartcard:
            QMessageBox.critical(
                None, _('Authentication'),
                "Authentication by client certificate stored on the "
                "smartcard is not supported")
            return EX_SOFTWARE

        if not all([CONF.cert, CONF.key, CONF.identity_url]):
            QMessageBox.critical(
                None, _('Authentication'),
                "One of required configuraton options "
                "are missing: cert, key, identity_url")
            return EX_SOFTWARE

        try:
            token = request_token(CONF.identity_url, CONF.cert, CONF.key)
        except ApplicationConnectionError as err:
            QMessageBox.critical(
                None, _('Authentication'), str(err))
            return EX_SOFTWARE
        if not token:
            QMessageBox.critical(
                None, _('Authentication'), "Authentication failed")
            return EX_SOFTWARE
        SESSION.token = token
        SESSION.token_auth = True
        window = InstanceRequestWindow()
        window.exec_()
        return EX_OK

    if CONF.use_smartcard:
        error_title = _('Smart card authentication')
        try:
            if not smartcard_api:
                raise smartcard_exc.SmartCardNotConfigured()

            session = smartcard_api.get_session()
            SESSION.smartcard_session = session
            atexit.register(session.closeSession)
            while True:
                pin_attempts = smartcard_api.get_pin_attempts()
                if pin_attempts == 0:
                    QMessageBox.critical(
                        None, _('Smart card authentication'),
                        _("PIN tries exceeded. Smart card is locked."))
                    return EX_SOFTWARE

                pin = PinWindow.get_pin(attempts=pin_attempts,
                                        min_len=session.min_pin_len,
                                        max_len=session.max_pin_len)
                if pin is None:
                    logger.warning('Smart card authentication aborted.')
                    return EX_SOFTWARE
                try:
                    username, password = smartcard_api.get_session_data(
                        session, pin.encode().decode('utf-8'))
                except IncorrectPin as err:
                    QMessageBox.critical(
                        None, error_title,
                        str(err))
                    continue
                break

            if not password:
                reason = _(
                    'No saved password found, please enter a new password.')
                password = PasswordWindow.change_password(reason=reason)
                if not password:
                    return EX_SOFTWARE

            SESSION.user = username
            SESSION.password = password
            SESSION.smartcard_used = True

        except smartcard_exc.CertificateNotFound:
            QMessageBox.critical(
                None, error_title,
                _('Unable to find any certificate on the smart card. '
                  'Please contact system administrator.'))
            return EX_SOFTWARE
        except smartcard_exc.AppropriateCertificateNotFound:
            QMessageBox.critical(
                None, error_title,
                _('Unable to find appropriate certificate '
                  'on the smart card. '
                  'Please contact system administrator.'))
            return EX_SOFTWARE
        except smartcard_exc.MultipleCertificatesFound:
            QMessageBox.critical(
                None, error_title,
                _('Found more than one appropriate certificate. '
                  'Please contact system administrator.'))
            return EX_SOFTWARE
        except smartcard_exc.SmartCardError:
            detailed_error(
                None, error_title,
                _("Environment is not configured. "
                  "Unable to get user login and password from smart card. "
                  "Please contact system administrator."),
                detailed=get_log_msg())
        else:
            window = InstanceRequestWindow()
            window.exec_()
            return EX_OK

    window = MainWindow()
    app.start(window, single=CONF.single_launch)
    return EX_OK


if __name__ == "__main__":
    sys.exit(main())
