# coding: utf-8
# pylint: disable=no-name-in-module, no-member

import base64
import logging
import random
import time
from collections import namedtuple

import requests

from tionix_vdi_client import CLIENT, CONNECTION_TYPE, exceptions
from tionix_vdi_client.helpers import (
    get_os_version, get_vdi_api_url, API_VERSION)
from tionix_vdi_client.i18n import ugettext as _, ungettext
from tionix_vdi_client.settings import CONF

logger = logging.getLogger(__name__)

UNEXPECTED_ERR_MSG = _("Unexpected error occurred.")
VM = namedtuple('VM', 'info status rdp_token')


class VdiClient(object):
    api_url = None

    def __init__(self, user=None, password=None, domain_name=None,
                 project=None, ip_address=None, token=None, token_auth=False):

        if token_auth:
            self.auth_params = {
                'token': token
            }
        else:
            self.auth_params = {
                'user': user,
                'password': password
            }

            if domain_name:
                self.auth_params['user_domain_name'] = domain_name

            if project:
                self.auth_params.update({
                    'auth_tenant': project,
                    'project_domain_name': domain_name,
                })

        self.client_params = {}
        if ip_address:
            self.client_params['ip_address'] = ip_address
        elif CONF.ikecfg:
            self.client_params['ip_address'] = CONF.ikecfg
        self.client_params.update({
            'conn_type': CONNECTION_TYPE,
            'os': get_os_version(),
            'client': CLIENT,
        })

        self.token = token
        self.wgsession = None
        self.rdp_token = self.generate_vdi_token()

    @property
    def config_api_urls(self):
        api_urls = [get_vdi_api_url(CONF.cloud)]
        if CONF.secondary_cloud:
            api_urls.append(get_vdi_api_url(CONF.secondary_cloud))
        if CONF.additional_clouds:
            api_urls += [get_vdi_api_url(url)
                         for url in CONF.additional_clouds]
        return api_urls

    def set_param(self, **kwargs):
        self.auth_params.update(kwargs)

    # pylint: disable=too-many-statements,inconsistent-return-statements
    def _action(self, action, body=None, stop_event=None, web_guard=False):
        headers = cookies = None
        if self.token:
            headers = {'Authorization': 'Bearer {}'.format(self.token)}
        if self.wgsession:
            cookies = {'WGSESSION': self.wgsession}

        api_urls = [self.api_url] if self.api_url else self.config_api_urls

        retries = CONF.retries
        if retries >= 0:
            retries += 1

        default_err_msg = exceptions.ErrorCodes.MESSAGES[
            exceptions.ErrorCodes.UNEXPECTED_ERROR]
        server_errs = {url: default_err_msg for url in api_urls}
        while retries > 0 or retries == -1:
            if stop_event and stop_event.is_set():
                raise exceptions.ThreadCancelError

            for api_url in api_urls:
                logger.info('Performing %s action on %s', action, api_url)

                if not web_guard:
                    url = '{0}/{1}/{2}/'.format(api_url, API_VERSION, action)
                else:
                    url = '{0}/{1}'.format(api_url, action)

                try:
                    logger.debug('Requesting url {0} with parameters: {1}'
                                 ''.format(url, body))
                    if body:
                        if not web_guard:
                            response = requests.post(url, headers=headers,
                                                     cookies=cookies,
                                                     json=body)
                        else:
                            response = requests.post(url, headers=headers,
                                                     cookies=cookies,
                                                     data=body)
                    else:
                        response = requests.get(url, cookies=cookies,
                                                headers=headers)
                except requests.exceptions.SSLError:
                    msg = _("Server SSL certificate verification failed.")
                    logger.error(msg)
                    server_errs[api_url] = msg
                except requests.exceptions.ConnectionError:
                    msg = _("Server is not available. "
                            "Check URL for connection.")
                    logger.error(msg)
                    server_errs[api_url] = msg
                except requests.exceptions.InvalidURL:
                    msg = _("Invalid URL. Check URL for connection.")
                    logger.error(msg)
                    server_errs[api_url] = msg
                else:
                    if web_guard:
                        self._parse_web_guard_response(response)
                        return
                    else:
                        try:
                            result = self._parse_response(response,
                                                          cloud=api_url)
                        except exceptions.VDIServerResponseError as err:
                            if err.code not in (
                                    exceptions.ErrorCodes.retriable_errors):
                                raise err
                            server_errs[api_url] = str(err)
                        else:
                            if result is not None:
                                self.api_url = api_url
                                return result

            if retries > 0:
                retries -= 1
                if retries > 0:
                    logger.info("Unable to connect to cloud. Tries left: {0}"
                                "".format(retries))
            elif retries < 0:
                logger.info("Unable to connect to cloud. Retry after {0} "
                            "seconds.".format(CONF.timeout))
            if retries != 0:
                logger.info(
                    _("{0} seconds timeout before reconnecting to the cloud.")
                    .format(CONF.timeout))
                if stop_event:
                    stop_event.wait(CONF.timeout)
                else:
                    time.sleep(CONF.timeout)

        msg = ungettext(
            "Unable to connect to server: {}",
            "Unable to connect to servers: {}",
            len(api_urls))
        fail_reason_msgs = [
            _("{server_url}\nReason: {err}").format(server_url=url, err=err)
            for url, err in list(server_errs.items())]
        raise exceptions.ApplicationConnectionError(
            msg.format("\n\n".join(fail_reason_msgs)))

    def web_guard_request(self, stop_event=None):
        """
        Requests session key from Web Guard
        """
        data = {
            'username': self.auth_params['user'],
            'password': self.auth_params['password'],
        }
        self._action('auth/api/login', stop_event=stop_event, body=data,
                     web_guard=True)

    def request_vm(self, stop_event=None):
        """
            Connects to server and retrieves instance IP address
            :returns: ip address
        """
        if CONF.web_guard and not self.wgsession:
            self.web_guard_request(stop_event=stop_event)
        data = dict(**self.auth_params)
        data.update(**self.client_params)
        response = self._action('request-vm', stop_event=stop_event, body=data)

        request_id = response.get('request_id')
        if not request_id:
            msg = _("Couldn't get request id from server response.")
            logger.error("Couldn't get request_id from json: %s", response)
            raise exceptions.VDIServerResponseError(msg)
        logger.info("Successfully connected to %s.", self.api_url)
        logger.debug("Got request_id: %s", request_id)
        return request_id

    def get_vm(self, request_id, stop_event=None):
        """
            Connect to server and retrieve vm info
            :returns: vm data
        """
        data = {'request': request_id}
        response = self._action('get-vm', stop_event=stop_event, body=data)
        info = response.get('server')
        status = response.get('status')
        self.rdp_token = response.get('vdi_token')

        return VM(info, status, self.rdp_token)

    def change_password(self, new_password, stop_event=None):
        if CONF.web_guard and not self.wgsession:
            self.web_guard_request(stop_event=stop_event)

        data = {
            'new_password': new_password
        }
        data.update(**self.auth_params)
        data.pop('auth_tenant', None)
        data.pop('project_domain_name', None)

        self._action('change-password', body=data, stop_event=stop_event)
        self.auth_params['password'] = new_password

    @staticmethod
    def _get_response_json(response):
        """Return json dict from response"""
        try:
            json_body = response.json()
        except ValueError:
            msg = _("Response parsing error.")
            logger.error(
                "Response doesn't have json. Request url: %s", response.url)
            raise exceptions.VDIServerResponseError(msg)

        logger.debug('Got following response: {0}'.format(json_body))
        return json_body

    def _parse_response(self, response, cloud=None):
        result = None
        self.token = response.headers.get('X-Auth-Token') or self.token

        if response.status_code in (500, 503):
            logger.error("Error %s. Url: %s."
                         "See VDIserver logs for more details.",
                         response.status_code, response.url)
        elif response.status_code >= 400:
            response_content = response.content.decode('utf-8')
            err_msg = "Error %s. Content: %s"
            logger.error(err_msg, response.status_code, response_content)

            if response.status_code == 400:
                error = exceptions.VDIServerResponseError.from_json(
                    response_content, cloud_url=cloud)
                raise error

        else:
            result = self._get_response_json(response)
        return result

    def _parse_web_guard_response(self, response):
        if response.status_code >= 400:
            response_content = response.content.decode('utf-8')
            err_msg = "Error %s. Content: %s"
            logger.error(err_msg, response.status_code, response_content)
            raise exceptions.WebGuardConnectionError()
        else:
            self.wgsession = response.cookies.get(
                'WGSESSION') or self.wgsession
            if not self.wgsession:
                raise exceptions.WebGuardCookieError()

    @staticmethod
    def generate_vdi_token():
        timestamp = str(int(time.time()))
        random_int = str(random.randint(0, 9))
        concatenated = int(timestamp + random_int)
        shifted = str(concatenated >> 1)
        return base64.b64encode(shifted.encode())
