# coding: utf-8
import argparse
import datetime
import io
import itertools
import logging
import os
import sys
from collections import defaultdict

import prettytable
import subprocess

# disable logger messages from imports
logging.disable(logging.INFO)

# pylint: disable=wrong-import-position
from tionix_vdi_client import EX_OK  # noqa
from tionix_vdi_client.helpers import suppress_stdout  # noqa
from tionix_vdi_client.i18n import ugettext as _  # noqa
from tionix_vdi_client.settings import get_log_file_path, CONF  # noqa
from tionix_vdi_client.tests.functional.runner import (
    ExtendedTextTestRunner, TestResultStatus)  # noqa

try:
    from tionix_vdi_client.version import __version__ as version
except ImportError:
    version = ''

logging.disable(logging.NOTSET)


logger = logging.getLogger(__name__)

MIN_ADJUSTABLE_WIDTH = 5
PADDING = 1

TEST_RES_LOG_FILE = os.path.join(
    os.path.dirname(get_log_file_path()),
    u'TIONIX.VDIclient_self_diagnostics_{0}.log'.format(datetime.date.today())
)


def adjust_columns_width(headers, rows, diff):
    columns_width = [0] * len(headers)
    for row in itertools.chain([headers], rows):
        for col_num, cell in enumerate(row):
            if not isinstance(cell, (unicode, str)):
                cell = unicode(cell)
            if isinstance(cell, str):
                cell = cell.encode().decode('utf-8')
            cell_width = max(map(len, cell.splitlines() + ['']))
            columns_width[col_num] = max(columns_width[col_num], cell_width)

    for i, width in reversed(list(enumerate(columns_width))):
        if width > MIN_ADJUSTABLE_WIDTH:
            new_width = max(width - diff, MIN_ADJUSTABLE_WIDTH)
            diff -= width - new_width
            columns_width[i] = new_width

    return columns_width


def build_table(headers, rows):
    table = prettytable.PrettyTable(headers)
    # prettytable doesn't support linebreaks in headers, so we add headers
    # to rows and print table without headers
    table.add_row(headers)
    for row in rows:
        table.add_row(row)
    table.header = False
    table.align = 'l'
    table.hrules = prettytable.ALL
    table.left_padding_width = table.right_padding_width = PADDING
    return table


def build_wrapped_table(headers=None, rows=()):
    """Build pretty table that nicely fits into user console window."""
    if not headers and rows:
        # when we don't have headers, first row is a header
        headers = rows[0]
        rows = rows[1:]

    table = build_table(headers, rows)
    table_width = len(str(table).splitlines()[0])

    columns = subprocess.getoutput('tput cols').strip()
    try:
        max_width = int(columns)
    except:
        max_width = 80

    if max_width and max_width < table_width:
        # On Windows, table should be slightly smaller
        # than console window to display properly.
        max_width -= 5

        table.left_padding_width = table.right_padding_width = 0
        diff = table_width - max_width - PADDING * len(headers) * 2
        if diff > 0:
            columns_width = adjust_columns_width(headers, rows, diff)
            for col_num, header in enumerate(headers):
                table.max_width[header] = columns_width[col_num]
    return table.get_string()


def calc_total_test_results(results):
    total_results = defaultdict(int)
    total_results['status'] = TestResultStatus.OK
    total_results['total_run'] = len(results)

    for result in results:
        status = result[1]
        total_results[status] += 1
        if status in (TestResultStatus.FAIL, TestResultStatus.ERROR):
            total_results['status'] = TestResultStatus.FAIL

    return total_results


def form_test_results_table(results, wrapped=False):
    """
    Forms pretty table with test results. If wrapped is True,
     table will be wrapped to fit console width.
    """

    headers = [
        _(u"Test name"),
        _(u"Status"),
        _(u"Failure reason"),
    ]

    if wrapped:
        table = build_wrapped_table(headers, results)
    else:
        table = build_table(headers, results).get_string()

    total_results = calc_total_test_results(results)

    test_res_output = _(
        u"TIONIX.VDIclient {version}\n"
        u"{table}\n"
        u"Ran {total_run} tests\n"
        u"{total_status} (success={succeeded}, failures={failed}, "
        u"errors={errors})"
    ).format(
        version=version,
        table=table,
        total_run=total_results['total_run'],
        total_status=total_results['status'],
        succeeded=total_results[TestResultStatus.OK],
        failed=total_results[TestResultStatus.FAIL],
        errors=total_results[TestResultStatus.ERROR],
    )

    return test_res_output


def build_variables_table(module_conf):
    result = ''
    if module_conf:
        headers = [u'N', u'Variable name', u'Value']
        rows = []
        for i, (variable, value) in enumerate(
                iter(sorted(module_conf.items())), 1):
            if isinstance(value, (list, tuple)):
                value = u", ".join(value)
            rows.append([unicode(i), variable, value])

        formatted_table = build_wrapped_table(headers, rows)

        table_name = _(u'TIONIX.VDIclient extracted variables')
        table_data = (
            u"{border}\n" u"{table_name}\n" u"{border}\n" u"{table_data}"
        ).format(table_name=table_name, border="=" * len(table_name),
                 table_data=formatted_table + '\n')
        result = "\n" + table_data
    return result


def log_test_result(log_input):
    """Writes given input to log file"""
    try:
        with io.open(TEST_RES_LOG_FILE, 'a', encoding='utf-8') as log:
            log.write(log_input + u"\n\n")
    except IOError as err:
        logger.error(
            _(u"Failed to save diagnostics result to %(file)s. %(err)s"),
            {'file': TEST_RES_LOG_FILE, 'err': err})
    sys.stdout.write(u"\n\n")
    sys.stdout.write(
        _(u"Result saved to {0}").format(TEST_RES_LOG_FILE)
    )
    sys.stdout.write(u"\n")


def run_diagnostics(debug=False):
    with suppress_stdout():
        # import here to avoid import tionix_client in main.py
        from tionix_vdi_client.tests.functional.run_tests import run_func_tests
        func_test_results = run_func_tests(runner=ExtendedTextTestRunner,
                                           debug=debug)
    console_output = form_test_results_table(func_test_results.tests_results,
                                             wrapped=True)
    variables_table = build_variables_table(CONF)
    console_output = variables_table + '\n' + console_output

    sys.stdout.write(console_output)

    log_output = form_test_results_table(func_test_results.tests_results)
    log_output = variables_table + '\n' + log_output
    log_test_result(log_output)
    return EX_OK


if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--debug', dest='debug', action='store_true')
    args = parser.parse_args()

    run_diagnostics(debug=args.debug)
    sys.stdout.write(_(u"Press ENTER to exit."))
    raw_input()
    sys.exit()
