# coding: utf-8
#  pylint: disable=no-member, no-name-in-module

import json
import logging
import string

from tionix_vdi_client import EX_SOFTWARE
from tionix_vdi_client.i18n import ugettext as _
from tionix_vdi_client.settings import CONF

logger = logging.getLogger(__name__)


if CONF.language == 'ru':
    SUPPORT_MESSAGE = CONF.contact_support_message_ru
else:
    SUPPORT_MESSAGE = CONF.contact_support_message_en


class ErrorCodes(object):
    UNEXPECTED_ERROR = 1
    LICENSE_ERROR = 2
    TASK_FAILED = 3
    TASK_TIME_OUT = 4
    NO_AVAILABLE_PROJECT = 5
    AUTH_ERROR = 6
    PASSWORD_EXPIRED = 7
    NON_VDI_PROJECT = 8
    INVALID_BODY = 9
    REQUEST_NOT_FOUND = 10
    REQUEST_ACCESS_FORBIDDEN = 11
    PROJECT_QUOTA_EXCEEDED = 12
    SERVER_CREATE_FAILED = 13
    NO_INSTANCE_IN_PROJECT = 14
    DOMAIN_NOT_FOUND = 15
    USER_NOT_FOUND = 16
    INSTANCE_STATE_ERROR = 17
    INSTANCE_NO_IP = 18
    PASSWORD_EXPIRE_WARNING = 19
    PASSWORD_ATTEMPTS_WARNING = 20
    KEYSTONE_SERVER_ERROR = 21
    KERBEROS_SCRIPT_FILE_PROBLEM = 22
    KERBEROS_SCRIPT_RETURN_1 = 23
    INVALID_TOKEN = 24
    TASK_PROCESSING_TIME_OUT = 25
    INAPPROPRIATE_WEB_GUARDED_REQUEST = 26

    retriable_errors = (
        UNEXPECTED_ERROR, LICENSE_ERROR,
        TASK_FAILED, TASK_TIME_OUT, TASK_PROCESSING_TIME_OUT,
        KEYSTONE_SERVER_ERROR,
        KERBEROS_SCRIPT_FILE_PROBLEM, KERBEROS_SCRIPT_RETURN_1
    )

    reset_token_errors = (
        PASSWORD_EXPIRED, PASSWORD_EXPIRE_WARNING, PASSWORD_ATTEMPTS_WARNING
    )

    MESSAGES = {
        UNEXPECTED_ERROR: _(
            "Unexpected error occurred. ") + SUPPORT_MESSAGE,
        LICENSE_ERROR: _(
            'License of VDI server {cloud_url} got out. ') + SUPPORT_MESSAGE,
        NO_AVAILABLE_PROJECT: _(
            'No available VDI project found. ') + SUPPORT_MESSAGE,
        AUTH_ERROR: _("Invalid credentials. ") + SUPPORT_MESSAGE,
        PASSWORD_EXPIRED: _("The password needs to be changed."),
        NON_VDI_PROJECT: _('Not a VDI project: {project_name}.'),
        PROJECT_QUOTA_EXCEEDED: _(
            "Quota exceeded for VDI project "
            "\"{project_name}\". ") + SUPPORT_MESSAGE,
        SERVER_CREATE_FAILED: _(
            'Unable to create instance in project {project_name}. '
        ) + SUPPORT_MESSAGE,
        NO_INSTANCE_IN_PROJECT: _(
            'No instance found in project {project_name}. '
        ) + SUPPORT_MESSAGE,
        DOMAIN_NOT_FOUND: _(
            'Domain {domain_name} not found.'),
        USER_NOT_FOUND: _(
            'User {user_name} not found in domain {domain_name}.'),
        INSTANCE_STATE_ERROR: _(
            'Instance with id {instance_id} is '
            'in invalid state: {instance_status}. ') + SUPPORT_MESSAGE,
        INSTANCE_NO_IP: _(
            'Unable to get IP address of instance with id {instance_id}. '
        ) + SUPPORT_MESSAGE,
        PASSWORD_EXPIRE_WARNING: _('Password expires in {seconds} seconds '
                                   'and needs to be changed.'),
        PASSWORD_ATTEMPTS_WARNING: _('Password expired, {attempts} grace '
                                     'logins remain.'),
        KEYSTONE_SERVER_ERROR: _(
            'Authentication server is not available. ') + SUPPORT_MESSAGE,
        KERBEROS_SCRIPT_FILE_PROBLEM: _(
            "Couldn't change password because of the KERBEROS.script_path "
            "file doesn't exist or it's unaccessible. ") + SUPPORT_MESSAGE,
        KERBEROS_SCRIPT_RETURN_1: _(
            "Couldn't change password because of an error while change "
            "password process execution. ") + SUPPORT_MESSAGE,
        # prevent to show original token message to user
        # show it in console log only
        INVALID_TOKEN: _("Unexpected error occurred. ") + SUPPORT_MESSAGE,
        INAPPROPRIATE_WEB_GUARDED_REQUEST: _("Invalid settings of Web Guard. "
                                             "") + SUPPORT_MESSAGE,
    }


class VDIException(Exception):
    pass


class VDIConnectionError(VDIException):
    pass


class ThreadCancelError(VDIException):
    pass


class VDIServerResponseError(VDIException):
    code = ErrorCodes.UNEXPECTED_ERROR

    def __init__(self, message, code=None, data=None):
        super(VDIServerResponseError, self).__init__(message)
        self._msg_original = message
        self.data = data or {}
        if code:
            self.code = code

    @classmethod
    def from_json(cls, json_str, cloud_url=None):
        error_obj = json.loads(json_str)
        error = error_obj.get('error', {})
        code = error.get('code')
        data = error.get('data')
        message = error.get('message')
        if data is None:
            data = {}
        data['cloud_url'] = cloud_url
        err_inst = cls(message, data=data)
        err_inst.code = code
        return err_inst

    def to_json(self):
        body = {
            'error': {
                'code': self.code,
                'data': self.data,
                'message': self.message
            }
        }
        return json.dumps(body)

    @property
    def message(self):
        msg = ErrorCodes.MESSAGES.get(self.code) or self._msg_original
        return self.safeformat(msg, **self.data)

    @staticmethod
    def safeformat(template, **kwargs):
        class SafeDict(dict):
            def __missing__(self, key):
                return ''

        return string.Formatter().vformat(template, (), SafeDict(**kwargs))


class ApplicationLogicException(VDIException):
    exit_code = EX_SOFTWARE


class MissingParameterException(VDIException):
    pass


class ApplicationConnectionError(VDIException):
    pass


class WebGuardConnectionError(VDIException):
    message = _("Unable to connect to Web Guard. ") + SUPPORT_MESSAGE


class WebGuardCookieError(VDIException):
    message = _("Unable to get session key from Web Guard. "
                "") + SUPPORT_MESSAGE
