# coding: utf-8
# pylint: disable=c-extension-no-member,no-name-in-module,import-error
import sys
import threading
from datetime import date

from PyQt5 import QtCore, QtGui
from PyQt5.QtWidgets import (
    QMainWindow, QDesktopWidget, QMessageBox, QPushButton, QSystemTrayIcon)

from tionix_vdi_client import MODULE_VERSION
from tionix_vdi_client.gui.request_window import InstanceRequestWindow
from tionix_vdi_client.gui.settings_window import (
    SettingsWindow, RDPSettingsWindow)
from tionix_vdi_client.gui.ui_main import Ui_MainWindow
from tionix_vdi_client.i18n import ugettext as _
from tionix_vdi_client.settings import SESSION, CONF


# noinspection PyUnresolvedReferences


# noinspection PyUnresolvedReferences
class MainWindow(QMainWindow):
    def __init__(self):
        super(MainWindow, self).__init__()
        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)

        # configure tray icon
        self.tray_icon = QSystemTrayIcon(self)
        icon = QtGui.QIcon()
        icon.addPixmap(
            QtGui.QPixmap(":/icons/favicon.gif"),
            QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.tray_icon.setIcon(icon)
        self.tray_icon.activated.connect(self.restore_from_tray)
        self.tray_icon.show()

        # disable maximize button
        self.setWindowFlags(
            self.windowFlags() & ~QtCore.Qt.WindowMaximizeButtonHint)

        self.ui.label_version.setText(
            self.ui.label_version.text().format(version=MODULE_VERSION))
        self.ui.label_copyright.setText(
            self.ui.label_copyright.text().format(year=date.today().year))

        self.center()

        self.ui.button_settings.clicked.connect(self.on_settings_clicked)

        self.ui.button_connect.clicked.connect(self.on_connect_clicked)
        self.ui.button_settings.installEventFilter(self)
        self.ui.button_exit.clicked.connect(self.close)
        self.ui.button_exit.installEventFilter(self)
        self.ui.checkbox_remember.stateChanged.connect(
            self.on_remember_changed)

        if SESSION.smartcard_used:
            self.ui.input_username.setEnabled(False)
            self.ui.input_password.setEnabled(False)

        self.ui.input_username.textChanged.connect(self.on_credentials_changed)
        self.ui.input_username.returnPressed.connect(self.on_connect_clicked)
        self.ui.input_password.textChanged.connect(self.on_credentials_changed)
        self.ui.input_password.returnPressed.connect(self.on_connect_clicked)

        self.ui.input_project.returnPressed.connect(self.on_connect_clicked)

        self.fill_credentials()

        if not CONF.project:
            self.ui.input_project.hide()
            self.ui.label_project.hide()
        if CONF.store_session:
            self.ui.checkbox_remember.setChecked(True)
        if not CONF.show_settings:
            self.ui.button_settings.hide()

        self.setFixedSize(self.ui.layout_main.sizeHint())

        self.on_credentials_changed()

    def eventFilter(self, obj, event):
        if (event.type() == QtCore.QEvent.KeyPress and
                isinstance(obj, QPushButton)):
            if (event.key() == QtCore.Qt.Key_Enter or
                    event.key() == QtCore.Qt.Key_Return):
                if obj.isEnabled():
                    obj.click()
                return True
        return super(MainWindow, self).eventFilter(obj, event)

    def fill_credentials(self):
        if SESSION.user:
            self.ui.input_username.setText(SESSION.user)
        if SESSION.password:
            self.ui.input_password.setText(SESSION.password)
        if CONF.project:
            if SESSION.project:
                self.ui.input_project.setText(SESSION.project)

    def center(self):
        qr = self.frameGeometry()
        cp = QDesktopWidget().availableGeometry().center()
        qr.moveCenter(cp)
        self.move(qr.topLeft())

    def on_settings_clicked(self):
        if sys.platform == 'win32':
            settings_dialog = RDPSettingsWindow(self)
        else:
            settings_dialog = SettingsWindow(self)
        settings_dialog.exec_()
        self.on_credentials_changed()

    def disable_connect_temporarly(self, interval=5):
        self.ui.button_connect.setEnabled(False)
        threading.Timer(
            interval=interval,
            function=lambda: self.ui.button_connect.setEnabled(True)
        ).start()

    def on_connect_clicked(self):
        if self.ui.button_connect.isEnabled():
            self.update_session()
            window = InstanceRequestWindow(self)
            window.exec_()
            self.fill_credentials()
            self.disable_connect_temporarly()

    def on_credentials_changed(self):
        required_params = self.get_auth_params()
        required_params.pop('project', None)
        connect_enabled = all(list(required_params.values()) + [CONF.cloud])
        self.ui.button_connect.setEnabled(connect_enabled)

    def is_session_changed(self):
        auth_params = self.get_auth_params()
        return not (
            SESSION.user == auth_params['user'] and
            SESSION.password == auth_params['password'] and
            SESSION.project == auth_params.get('project', '') and
            SESSION.domain_name == auth_params['domain_name'])

    def update_session(self):
        auth_params = self.get_auth_params()
        SESSION.user = auth_params['user']
        SESSION.password = auth_params['password']
        SESSION.project = auth_params.get('project', '')
        SESSION.domain_name = auth_params['domain_name']
        SESSION.save()

    def on_remember_changed(self):
        CONF.store_session = self.ui.checkbox_remember.isChecked()
        CONF.save()

    def closeEvent(self, event):
        if CONF.store_session:
            if self.is_session_changed():
                ret = QMessageBox.question(
                    self,
                    _('Exit'),
                    _('Connection settings are not saved. '
                      'Would you like to save them before exit?'),
                    buttons=(
                        QMessageBox.Save |
                        QMessageBox.Discard |
                        QMessageBox.Cancel
                    )
                )
                if ret == QMessageBox.Save:
                    self.update_session()
                elif ret == QMessageBox.Cancel:
                    event.ignore()
                    return

        ret = QMessageBox.question(
            self, _('Exit'), _("Are you sure you want to quit?"),
            QMessageBox.Yes, QMessageBox.No)

        if ret == QMessageBox.Yes:
            event.accept()
            QtCore.QCoreApplication.quit()
        else:
            event.ignore()

    def get_auth_params(self):
        auth_params = {
            'user': self.ui.input_username.text(),
            'password': self.ui.input_password.text(),
            'project': self.ui.input_project.text(),
            'domain_name': CONF.domain_name,
        }
        if not CONF.project:
            auth_params.pop('project')
        return auth_params

    def hide_to_tray(self):
        self.hide()
        self.setWindowFlags(self.windowFlags() & ~QtCore.Qt.Tool)

    def restore_from_tray(self):
        self.showNormal()
        self.activateWindow()
