# coding: utf-8
# pylint: disable=c-extension-no-member,no-name-in-module,import-error
import logging

from PyQt5 import QtGui
from PyQt5.QtWidgets import QDialog, QMessageBox, QLineEdit

from tionix_vdi_client import exceptions
from tionix_vdi_client.client import VdiClient
from tionix_vdi_client.gui.messagebox import detailed_error
from tionix_vdi_client.gui.ui_password import Ui_PasswordWindow
from tionix_vdi_client.helpers import PasswordGenerator
from tionix_vdi_client.i18n import ugettext as _
from tionix_vdi_client.log import get_log_msg
from tionix_vdi_client.settings import SESSION, CONF
from tionix_vdi_client.smart_card import api as smartcard_api

logger = logging.getLogger(__name__)


class PasswordWindow(QDialog):
    new_password = None
    eye_opened = ":/icons/eye_opened.png"
    eye_crossed = ":/icons/eye_crossed.png"

    def __init__(self, parent=None, reason=None):
        super(PasswordWindow, self).__init__(parent)
        self.ui = Ui_PasswordWindow()
        self.ui.setupUi(self)

        if reason:
            self.ui.label_reason.setText(reason)
        else:
            self.ui.label_reason.setVisible(False)

        self.init_password_inputs()

        self.ui.input_user.setText(SESSION.user)
        self.ui.input_current_password.setText(SESSION.password)

        self.ui.input_current_password.textChanged.connect(
            self.on_input_change)
        self.ui.input_current_password.returnPressed.connect(
            self.on_save_clicked)

        self.ui.input_new_password.textChanged.connect(self.on_input_change)
        self.ui.input_new_password.returnPressed.connect(self.on_save_clicked)

        self.new_password_eye = self.ui.input_new_password.addAction(
            QtGui.QIcon(self.eye_opened),
            QLineEdit.ActionPosition.TrailingPosition)
        self.new_password_eye.triggered.connect(self.change_new_password_state)

        self.ui.input_confirm_password.textChanged.connect(
            self.on_input_change)
        self.ui.input_confirm_password.returnPressed.connect(
            self.on_save_clicked)

        self.confirm_password_eye = self.ui.input_confirm_password.addAction(
            QtGui.QIcon(self.eye_opened),
            QLineEdit.ActionPosition.TrailingPosition)
        self.confirm_password_eye.triggered.connect(
            self.change_confirm_password_state)

        self.ui.button_save.clicked.connect(self.on_save_clicked)
        self.ui.button_cancel.clicked.connect(self.on_cancel_clicked)
        self.on_input_change()

        self.ui.chekbox_own_password.stateChanged.connect(
            self._password_state_change)

    @staticmethod
    def _get_bg_color(widget):
        rgb = widget.palette().background().color().toRgb()
        r, g, b = rgb.red(), rgb.green(), rgb.blue()
        return r, g, b

    def _disable_field(self, widget):
        widget.setReadOnly(True)
        widget.setStyleSheet(
            'QLineEdit {background-color: rgb(%s, %s, %s); '
            'color: rgb(190, 190, 190)}' % self._get_bg_color(self))

    def _password_state_change(self):
        if self.ui.input_new_password.isReadOnly():
            self.ui.input_new_password.setReadOnly(False)
            self.ui.input_confirm_password.setReadOnly(False)
            style = ('QLineEdit {background-color: rgb(255, 255, 255); '
                     'color: rgb(0, 0, 0)}')
            self.ui.input_new_password.setStyleSheet(style)
            self.ui.input_confirm_password.setStyleSheet(style)
        else:
            self._disable_field(self.ui.input_new_password)
            self._disable_field(self.ui.input_confirm_password)

    def init_password_inputs(self):
        if CONF.password_generation:
            random_password = PasswordGenerator.generate()
            self.ui.input_new_password.setText(random_password)
            self.ui.input_confirm_password.setText(random_password)

            self._disable_field(self.ui.input_new_password)
            self._disable_field(self.ui.input_confirm_password)
        else:
            self.ui.chekbox_own_password.setVisible(False)

    def change_new_password_state(self):
        self.change_state(self.ui.input_new_password, self.new_password_eye)

    def change_confirm_password_state(self):
        self.change_state(self.ui.input_confirm_password,
                          self.confirm_password_eye)

    def change_state(self, widget, action):
        if widget.echoMode() is QLineEdit.Password:
            widget.setStyleSheet(
                'QLineEdit {background-color: rgb(%s, %s, %s); '
                'color: rgb(0, 0, 0)}' % self._get_bg_color(widget))
            action.setIcon(QtGui.QIcon(self.eye_crossed))
            widget.setEchoMode(QLineEdit.Normal)
        else:
            if widget.isReadOnly():
                widget.setStyleSheet(
                    'QLineEdit {background-color: rgb(%s, %s, %s); color: '
                    'rgb(190, 190, 190)}' % self._get_bg_color(widget))
            else:
                widget.setStyleSheet(
                    'QLineEdit {background-color: rgb(%s, %s, %s); color: '
                    'rgb(0, 0, 0)}' % self._get_bg_color(widget))
            action.setIcon(QtGui.QIcon(self.eye_opened))
            widget.setEchoMode(QLineEdit.Password)

    def on_input_change(self):
        current_password = self.ui.input_current_password.text()
        new_password = self.ui.input_new_password.text()
        confirm_password = self.ui.input_confirm_password.text()
        enabled = False
        if current_password and new_password and confirm_password:
            enabled = True
        self.ui.button_save.setEnabled(enabled)

    def on_save_clicked(self):
        if not self.ui.button_save.isEnabled():
            return

        current_password = self.ui.input_current_password.text()
        new_password = self.ui.input_new_password.text()
        confirm_new_password = self.ui.input_confirm_password.text()

        if new_password != confirm_new_password:
            msg = _('Passwords do not match. Make sure they are the same.')
            QMessageBox.information(self, _('Error'), msg)
            return

        if CONF.password_generation:
            if not PasswordGenerator.check(new_password):
                msg = _("The password must contain latin letters in upper "
                        "(A-Z) and lower (a-z) case, at least 1 digit (0-9) "
                        "and at least 1 special character (!@#$%^&*).")
                QMessageBox.information(self, _('Error'), msg)
                self.init_password_inputs()
                return

        # TODO: move APIrequests to QThread to prevent ui blocking
        vdi_client = VdiClient(
            user=SESSION.user,
            password=current_password,
            domain_name=CONF.domain_name,
            project=SESSION.project,
            ip_address=CONF.ikecfg,
            token=SESSION.token)

        try:
            vdi_client.change_password(new_password)
        except (exceptions.VDIServerResponseError,
                exceptions.WebGuardCookieError,
                exceptions.WebGuardConnectionError) as err:
            detailed_error(self, _('Connection error'), str(err),
                           detailed=get_log_msg())
            self.init_password_inputs()
        else:
            self.new_password = new_password

            if SESSION.smartcard_session:
                result = smartcard_api.update_password(
                    SESSION.smartcard_session, SESSION.user, new_password)
                if not result:
                    logger.warning('Error while updating smartcard password.')

            QMessageBox.information(
                self, '', _('Password successfully changed.'))

            # close only on successful password change
            self.close()

    def on_cancel_clicked(self):
        self.reject()

    @staticmethod
    def change_password(parent=None, reason=None):
        window = PasswordWindow(parent, reason=reason)
        window.exec_()
        return window.new_password
