# coding: utf-8
# pylint: disable=c-extension-no-member,no-name-in-module,import-error
import json
import logging
import threading

from PyQt5 import QtCore
from PyQt5.QtWidgets import QDialog, QMessageBox

from tionix_vdi_client import exceptions
from tionix_vdi_client.client import VdiClient
from tionix_vdi_client.gui.messagebox import detailed_error
from tionix_vdi_client.gui.password_window import PasswordWindow
from tionix_vdi_client.gui.ui_request import Ui_InstanceRequestWindow
from tionix_vdi_client.helpers import RequestStatus, ServerTaskStates
from tionix_vdi_client.i18n import ugettext as _
from tionix_vdi_client.log import get_log_msg
from tionix_vdi_client.rdp import rdp_connect
from tionix_vdi_client.settings import CONF, SESSION

logger = logging.getLogger(__name__)


REQUEST_MSG = _('Request instance for {user}')
REQUEST_PENDING_MSG = _('Pending request processing...')
REQUEST_PROCESS_MSG = _('Processing request...')
ACTION_PERFOMED_MSG = _("{0} is performed for user's instance.")


# noinspection PyUnresolvedReferences
class RdpConnectionThread(QtCore.QThread):

    # OUT signals
    connection_error = QtCore.pyqtSignal(str)
    connection_success = QtCore.pyqtSignal()

    # IN signals
    stop_requested = QtCore.pyqtSignal()

    def __init__(self):
        super(RdpConnectionThread, self).__init__()
        self.stop_event = threading.Event()
        self.stop_requested.connect(self.stop)

    def stop(self):
        self.stop_event.set()

    def run(self):
        try:
            rdp_connect(
                user=SESSION.user,
                password=SESSION.password,
                domain_name=CONF.domain_name,
                vm_ip=SESSION.ip,
                rdp_token=SESSION.rdp_token,
                stop_event=self.stop_event,
                silent=CONF.silent)
        except Exception as err:
            self.connection_error.emit(str(err))
        else:
            self.connection_success.emit()


# noinspection PyUnresolvedReferences
class InstanceRequestThread(QtCore.QThread):

    # OUT signals
    instance_received = QtCore.pyqtSignal(str)
    state_updated = QtCore.pyqtSignal(str)
    response_error = QtCore.pyqtSignal(str)
    request_error = QtCore.pyqtSignal(str)

    # IN signals
    stop_requested = QtCore.pyqtSignal()

    # TODO: create vdi_client in thread, pass auth parameters
    def __init__(self, vdi_client):
        super(InstanceRequestThread, self).__init__()
        self.vdi_client = vdi_client
        self.stop_event = threading.Event()
        self.stop_requested.connect(self.stop)

    def stop(self):
        self.stop_event.set()

    def run(self):
        cancel_message = 'Connection thread was canceled.'
        try:
            request_id = self.vdi_client.request_vm(stop_event=self.stop_event)
            while not self.stop_event.is_set():
                vm_request = self.vdi_client.get_vm(
                    request_id, stop_event=self.stop_event)
                vm_request_json = json.dumps({
                    'info': vm_request.info or {},
                    'status': vm_request.status,
                    'rdp_token': vm_request.rdp_token})
                if vm_request.status == RequestStatus.COMPLETED:
                    self.instance_received.emit(vm_request_json)
                    break
                else:
                    self.state_updated.emit(vm_request_json)

                self.stop_event.wait(CONF.get_vm_timeout)
            else:
                logger.info(cancel_message)
        except exceptions.ThreadCancelError:
            logger.info(cancel_message)
        except exceptions.VDIServerResponseError as err:
            self.response_error.emit(err.to_json())
        except (exceptions.WebGuardConnectionError,
                exceptions.WebGuardCookieError) as err:
            exc = exceptions.VDIServerResponseError(
                str(err), code=exceptions.ErrorCodes.AUTH_ERROR)
            self.response_error.emit(exc.to_json())
        except Exception as err:
            message = getattr(err, 'message', str(err))
            self.request_error.emit(message)


class InstanceRequestWindow(QDialog):

    instance_ip = None
    vdi_client = None
    request_thread = None
    connection_thread = None

    def __init__(self, parent=None):
        super(InstanceRequestWindow, self).__init__(parent)
        self._parent = parent
        self.ui = Ui_InstanceRequestWindow()
        self.ui.setupUi(self)

        self.ui.label_details.setText(REQUEST_MSG.format(user=SESSION.user))
        self.ui.button_cancel.clicked.connect(self.close)
        self.start_request_thread()

    def start_request_thread(self):
        if getattr(SESSION, 'token_auth'):
            auth_params = dict(
                token=SESSION.token,
                token_auth=True)
        else:
            auth_params = dict(
                user=SESSION.user,
                password=SESSION.password,
                domain_name=CONF.domain_name,
                project=SESSION.project,
            )

        self.vdi_client = VdiClient(
            ip_address=CONF.ikecfg, **auth_params)

        self.set_progress_animation(active=True)
        self.request_thread = InstanceRequestThread(self.vdi_client)
        self.request_thread.instance_received.connect(
            self.on_instance_received)
        self.request_thread.state_updated.connect(self.on_state_updated)
        self.request_thread.response_error.connect(self.on_response_error)
        self.request_thread.request_error.connect(self.on_request_error)
        self.request_thread.start()

    def start_connection_thread(self):
        self.connection_thread = RdpConnectionThread()
        self.connection_thread.connection_error.connect(
            self.on_connection_error)
        self.connection_thread.connection_success.connect(
            self.on_connection_success)
        if CONF.minimize_to_tray and self._parent:
            self.set_progress_animation(active=False)
            self.hide()
            self._parent.hide_to_tray()
        self.connection_thread.start()

    def set_progress_animation(self, active=True):
        if active:
            self.ui.progress_bar.setMaximum(0)
        else:
            self.ui.progress_bar.setMaximum(1)

    def on_instance_received(self, vm_request_json):
        vm_request = json.loads(vm_request_json)
        info = vm_request.get('info')
        ip = info.get('ip')
        SESSION.ip = ip
        SESSION.rdp_token = vm_request.get('rdp_token')
        logger.debug('Got instance ip: %s', ip)
        self.ui.label_details.setText(
            _('Connecting to instance {}...').format(ip))
        self.start_connection_thread()

    def on_state_updated(self, vm_request_json):
        vm_request = json.loads(vm_request_json)
        status = vm_request.get('status')
        info = vm_request.get('info')
        state_details = None

        if status == RequestStatus.PENDING:
            state_details = REQUEST_PENDING_MSG

        elif status == RequestStatus.PROCESS:
            state_details = REQUEST_PROCESS_MSG
            server_task_state = info.get('task_state')

            if server_task_state in list(ServerTaskStates.keys()):
                state_details = ACTION_PERFOMED_MSG.format(
                    ServerTaskStates[server_task_state])
        if state_details:
            self.ui.label_details.setText(state_details)

    def on_response_error(self, error_json):
        self.set_progress_animation(False)
        err = exceptions.VDIServerResponseError.from_json(error_json)

        change_pwd = False
        reason = _('Invalid credentials')

        if err.code in exceptions.ErrorCodes.reset_token_errors:
            reason = str(err)
            change_pwd = True
        elif err.code == exceptions.ErrorCodes.AUTH_ERROR:
            if SESSION.smartcard_used:
                reply = QMessageBox.question(
                    self, reason,
                    _('If you get temporary password, you could change '
                      'it to a permanent one. Do you want change it now?'),
                    QMessageBox.Yes, QMessageBox.No)
                change_pwd = reply == QMessageBox.Yes
            else:
                detailed_error(self, reason, str(err),
                               detailed=get_log_msg())
        else:
            detailed_error(self, _('Connection error'), str(err),
                           detailed=get_log_msg())

        if change_pwd:
            SESSION.token = self.vdi_client.token
            new_password = PasswordWindow.change_password(self, reason=reason)

            if new_password:
                SESSION.password = new_password
                self.start_request_thread()
                return

        self.request_thread.wait()
        self.close()

    def on_request_error(self, error_msg):
        self.set_progress_animation(False)
        detailed_error(self, _('Connection error'), error_msg,
                       detailed=get_log_msg())
        self.request_thread.wait()
        self.close()

    def on_connection_error(self, error_msg):
        self.set_progress_animation(active=False)
        logger.error('Connection reset: %s', error_msg)
        self.connection_thread.wait()
        self.close()

    def on_connection_success(self):
        self.set_progress_animation(active=False)
        self.connection_thread.wait()
        self.close()

    def closeEvent(self, event):
        interrupt = True
        if self.request_thread and self.request_thread.isRunning():
            reply = QMessageBox.question(
                self, _('Exit'),
                _("Are you sure you want to interrupt the waiting "
                  "for response on instance request?"),
                QMessageBox.Yes, QMessageBox.No)

            if reply != QMessageBox.Yes:
                interrupt = False
        elif self.connection_thread and self.connection_thread.isRunning():
            reply = QMessageBox.question(
                self, _('Exit'),
                _("Are you sure you want to interrupt RDP connection?"),
                QMessageBox.Yes, QMessageBox.No)

            if reply != QMessageBox.Yes:
                interrupt = False

        if interrupt:
            if self.request_thread and self.request_thread.isRunning():
                self.request_thread.stop_requested.emit()
                self.request_thread.wait()
            if self.connection_thread and self.connection_thread.isRunning():
                self.connection_thread.stop_requested.emit()
                self.connection_thread.wait()
            event.accept()
        else:
            event.ignore()

    @staticmethod
    def request_instance(parent=None):
        window = InstanceRequestWindow(parent)
        window.exec_()
        return window.instance_ip
