# coding: utf-8
# pylint: disable=c-extension-no-member,no-name-in-module,import-error
import sys
import logging

from PyQt5 import QtCore
from PyQt5.QtGui import QStandardItem, QStandardItemModel
from PyQt5.QtWidgets import QDialog, QMessageBox

from tionix_vdi_client import Localization
from tionix_vdi_client.gui.ui_settings import Ui_SettingsWindow
from tionix_vdi_client.i18n import ugettext as _
# noinspection PyProtectedMember
from tionix_vdi_client.rdp import (
    read_rdp_config, save_rdp_config, _parse_config)
from tionix_vdi_client.settings import CONF, update_logging_level
from tionix_vdi_client.settings import RDP_FILE_SAMPLE_PATH
from tionix_vdi_client.win.devices import (
    supported_device_classes, get_pnp_devices,
    find_parent_device_id, find_device_children_ids)
from tionix_vdi_client.win.drives import get_drives, get_display_name


logger = logging.getLogger(__name__)


# noinspection PyUnresolvedReferences
class SettingsWindow(QDialog):
    def __init__(self, parent=None):
        super(SettingsWindow, self).__init__(parent)
        self.ui = Ui_SettingsWindow()
        self.ui.setupUi(self)

        self.ui.button_cancel.clicked.connect(self.reject)
        self.ui.button_save.clicked.connect(self.on_save_clicked)

        if CONF.cloud:
            self.ui.input_primary_server.setText(CONF.cloud)
        self.ui.input_primary_server.returnPressed.connect(
            self.on_save_clicked)
        if CONF.secondary_cloud:
            self.ui.input_secondary_server.setText(CONF.secondary_cloud)
        self.ui.input_secondary_server.returnPressed.connect(
            self.on_save_clicked)
        if CONF.domain_name:
            self.ui.input_domain.setText(CONF.domain_name)
        self.ui.input_domain.returnPressed.connect(self.on_save_clicked)

        self.ui.combobox_language.setCurrentText(
            Localization.locales_translate.get(CONF.language))
        self.ui.combobox_log_level.setCurrentText(CONF.log_level)

        if sys.platform != "win32":
            self.ui.tab_widget.removeTab(1)

    def _extra_save(self):
        pass

    def on_save_clicked(self):
        if self.ui.button_save.isEnabled():
            old_language = CONF.language
            CONF.cloud = self.ui.input_primary_server.text()
            CONF.secondary_cloud = self.ui.input_secondary_server.text()
            CONF.domain_name = self.ui.input_domain.text()
            CONF.language = Localization.get_code_by_locale(
                self.ui.combobox_language.currentText())
            CONF.log_level = self.ui.combobox_log_level.currentText()
            CONF.save()
            update_logging_level()
            self._extra_save()
            if CONF.language != old_language:
                QMessageBox.information(
                    self, 'TIONIX.VDIclient',
                    _('TIONIX.VDIclient should be restarted to '
                      'change localization.'),
                    QMessageBox.Ok)
            self.close()


# noinspection PyUnresolvedReferences
class RDPSettingsWindow(SettingsWindow):
    def __init__(self, parent=None):
        super(RDPSettingsWindow, self).__init__(parent)

        self.tree_model = QStandardItemModel()
        self.ui.tree_devices.setModel(self.tree_model)
        self.ui.tree_devices.setHeaderHidden(True)
        self.tree_model.setColumnCount(1)

        self._drives = []
        self._devices = []
        self._devices_root = None
        self._drives_root = None

        self._logical_drives = get_drives()
        self._pnp_devices = get_pnp_devices()
        rdp_conf = _parse_config(read_rdp_config(RDP_FILE_SAMPLE_PATH))
        self.ui.input_ldap_domain.setText(rdp_conf['domain'])
        self.ui.checkbox_multiple_monitors.setChecked(
            rdp_conf['use_multimon'])
        self.init_devices(
            self._logical_drives, self._pnp_devices,
            rdp_conf['drives'], rdp_conf['devices'],
            rdp_conf['other_devices'])

    def init_devices(self, logical_drives, pnp_devices,
                     selected_drives, selected_devices,
                     other_devices=False):

        selected_drives = selected_drives or []
        selected_devices = selected_devices or []
        parsed_selected_devices = []
        for device_id in selected_devices:
            children = find_device_children_ids(device_id, self._pnp_devices)
            if children:
                parsed_selected_devices.extend(children)
            else:
                parsed_selected_devices.append(device_id)

        all_drives = '*' in selected_drives
        other_drives = 'DynamicDrives' in selected_drives or all_drives
        self._drives = []
        drives_root = QStandardItem(_("Drives"))
        for drive in logical_drives:
            # store drives and get by index later
            drive_name = get_display_name(drive)
            self._drives.append(drive_name)
            child = QStandardItem(drive_name)
            child.setCheckable(True)
            if drive_name in selected_drives or all_drives:
                child.setCheckState(QtCore.Qt.Checked)
            drives_root.appendRow(child)
        child = QStandardItem(_("Drives that I plug in later"))
        child.setCheckable(True)
        if other_drives:
            child.setCheckState(QtCore.Qt.Checked)
        drives_root.appendRow(child)
        self.tree_model.appendRow(drives_root)

        self._devices = []
        devices_root = QStandardItem(
            _("Other supported Plug and Play (PnP) devices"))
        supported_classes = supported_device_classes()
        for device in pnp_devices:
            if device.ClassGuid not in supported_classes:
                continue
            # store devices and get by index later
            self._devices.append(device.DeviceID)
            child = QStandardItem(
                "{} ({})".format(device.Name.encode().decode("utf-8"), device.DeviceID))
            devices_root.appendRow(child)
            child.setCheckable(True)
            if device.DeviceID in parsed_selected_devices:
                child.setCheckState(QtCore.Qt.Checked)

        child = QStandardItem(_("Devices that I plug in later"))
        child.setCheckable(True)
        if other_devices:
            child.setCheckState(QtCore.Qt.Checked)
        devices_root.appendRow(child)
        self.tree_model.appendRow(devices_root)
        self._devices_root = devices_root
        self._drives_root = drives_root

    @staticmethod
    def _save_rdp_settings(domain, drives, devices,
                           other_devices=False, use_multimon=False):
        content = []
        conf = [
            'domain:s:{}\r\n'.format(domain),
            'use multimon:i:{}\r\n'.format(int(use_multimon)),
            'usbdevicestoredirect:s:{}\r\n'.format(';'.join(devices)),
            'devicestoredirect:s:{}\r\n'.format(
                '*' if other_devices else ''),
            'drivestoredirect:s:{}\r\n'.format(';'.join(drives))
        ]
        updated_keys = [line.split(':')[0] for line in conf]
        for line in read_rdp_config(RDP_FILE_SAMPLE_PATH):
            key = str(line.split(':')[0])
            if key in updated_keys:
                continue
            content.append(line)

        for line in conf:
            content.append(line)

        save_rdp_config(RDP_FILE_SAMPLE_PATH, content)

    def _extra_save(self):
        selected_drives = []
        # TODO: save to rdp.conf
        for i in range(self._drives_root.rowCount() - 1):
            child = self._drives_root.child(i)
            if child.checkState() == QtCore.Qt.Checked:
                device_id = self._drives[i]
                selected_drives.append(device_id)
        other_drives_item = self._drives_root.child(
            self._drives_root.rowCount() - 1)
        other_drives = (
            other_drives_item.checkState() == QtCore.Qt.Checked)
        if other_drives:
            selected_drives.append('DynamicDrives')
        if all(drive in selected_drives
               for drive in self._drives) and other_drives:
            selected_drives = ['*']

        selected_devices = []
        for i in range(self._devices_root.rowCount() - 1):
            child = self._devices_root.child(i)
            if child.checkState() == QtCore.Qt.Checked:
                device_id = self._devices[i]
                parent_id = find_parent_device_id(
                    device_id, self._pnp_devices)
                device_id = parent_id or device_id
                selected_devices.append(device_id)
        other_devices_item = self._devices_root.child(
            self._devices_root.rowCount() - 1)
        other_devices = (
            other_devices_item.checkState() == QtCore.Qt.Checked)
        self._save_rdp_settings(
            domain=self.ui.input_ldap_domain.text(),
            drives=selected_drives,
            devices=selected_devices,
            other_devices=other_devices,
            use_multimon=(
                self.ui.checkbox_multiple_monitors.isChecked())
        )
