# coding: utf-8

import logging
import platform
import re
import sys
import urllib.parse
from io import StringIO
from contextlib import contextmanager
from random import SystemRandom

from tionix_vdi_client.i18n import ugettext as _

logger = logging.getLogger(__name__)


API_VERSION = 'v1'
API_DEFAULT_PORT = '9365'
ALLOWED_SCHEMES = ('http://', 'https://')


def get_win_info():
    return '{system} {version}'.format(
        system=platform.system(),
        version=platform.release())


def get_linux_info():
    distname, version, _id = platform.linux_distribution()
    return '{distname} {version}'.format(
        distname=distname,
        version=version)


def get_os_version():
    """Get operating system version"""
    system = platform.system()
    info_getters = {
        'Linux': get_linux_info,
        'Windows': get_win_info,
    }
    info_getter = info_getters.get(system, platform.platform)
    return info_getter()


@contextmanager
def suppress_stdout():
    new_stdout = StringIO()
    old_stdout = sys.stdout
    old_stderr = sys.stderr
    logging.disable(logging.ERROR)
    try:
        sys.stdout = new_stdout
        sys.stderr = new_stdout
        yield
    finally:
        sys.stdout = old_stdout
        sys.stderr = old_stderr
        logging.disable(logging.NOTSET)


def get_vdi_api_url(url):
    """Get VDIserver API url

    Transforms url as follows:
    test.stand.loc -> http://test.stand.loc:9365
    test.stand.loc:80 -> http://test.stand.loc:80
    test.stand.loc:80/vdi/api -> http://test.stand.loc:80
    http://test.stand.loc -> http://test.stand.loc:9365
    test.stand.loc:9365 -> http://test.stand.loc:9365
    http://test.stand.loc/vdi/api -> http://test.stand.loc:9365
    """
    if not any(url.startswith(scheme) for scheme in ALLOWED_SCHEMES):
        url = ALLOWED_SCHEMES[0] + url

    scheme, netloc, _path, _query, _fragment = urllib.parse.urlsplit(url)

    netloc_parsed = netloc.split(':')
    if len(netloc_parsed) == 1:
        netloc_parsed.append(API_DEFAULT_PORT)
    netloc = ':'.join(netloc_parsed)

    vdi_api_url = urllib.parse.urlunsplit((scheme, netloc, '', '', ''))
    return vdi_api_url


class RequestStatus(object):

    PENDING = 'pending'
    PROCESS = 'process'
    ERROR = 'error'
    COMPLETED = 'completed'

    values = (
        PENDING,
        PROCESS,
        ERROR,
        COMPLETED,
    )

    READY_STATUSES = (ERROR, COMPLETED)


class PasswordGenerator(object):
    PASSWORD_MIN = 8

    LOWER = "abcdefghijklmnopqrstuvwxyz"
    UPPER = LOWER.upper()
    DIGITS = "0123456789"
    SYMBOLS = "!@#$%^&*"
    LOWER_MIN = UPPER_MIN = DIGITS_MIN = SYMBOLS_MIN = 1

    PASSWORD_PATTERN = re.compile(r'^[a-zA-Z0-9!@#$%^&*]+$')
    UPPER_PATTERN = re.compile(r'[A-Z]')
    DIGIT_PATTERN = re.compile(r'[0-9]')
    SYMBOLS_PATTERN = re.compile(r'[%s]' % SYMBOLS)

    @classmethod
    def generate(cls, password_len=None):
        random = SystemRandom()  # use system generator
        password_len = password_len or cls.PASSWORD_MIN
        parts = []

        upper = random.randint(cls.UPPER_MIN,
                               (password_len - cls.DIGITS_MIN -
                                cls.SYMBOLS_MIN - cls.LOWER_MIN))
        password_len -= upper
        parts.append((upper, cls.UPPER))
        digits = random.randint(cls.DIGITS_MIN,
                                password_len - cls.SYMBOLS_MIN - cls.LOWER_MIN)
        password_len -= digits
        parts.append((digits, cls.DIGITS))
        symbols = random.randint(cls.SYMBOLS_MIN, password_len - cls.LOWER_MIN)
        parts.append((symbols, cls.SYMBOLS))
        lower = password_len - symbols
        parts.append((lower, cls.LOWER))

        password = []
        for n, alphabet in parts:
            for _i in range(n):
                password.append(random.choice(alphabet))
        random.shuffle(password)

        return ''.join(password)

    @classmethod
    def check(cls, password):
        valid = False
        if not (len(password) < cls.PASSWORD_MIN or
                not cls.PASSWORD_PATTERN.match(password) or
                cls.UPPER_MIN and not cls.UPPER_PATTERN.search(password) or
                cls.DIGITS_MIN and not cls.DIGIT_PATTERN.search(password) or
                cls.SYMBOLS_MIN and not cls.SYMBOLS_PATTERN.search(password)):
            valid = True
        return valid


ServerTaskStates = {
    'block_device_mapping': _('Block device mapping'),
    'deleting': _('Deleting'),
    'force_delete': _('Force delete'),
    'networking': _('Networking'),
    'pausing': _('Pausing'),
    'powering-off': _('Powering off'),
    'powering-on': _('Powering on'),
    'rebooting': _('Rebooting'),
    'rebuilding': _('Rebuilding'),
    'rescuing': _('Rescuing'),
    'resize_confirming': _('Resize confirming'),
    'resize_finish': _('Resize finish'),
    'resize_migrated': _('Resize migrated'),
    'resize_migrating': _('Resize migrating'),
    'resize_prep': _('Resize prep'),
    'resize_reverting': _('Resize reverting'),
    'resuming': _('Resuming'),
    'scheduling': _('Scheduling'),
    'shelving': _('Shelving'),
    'spawning': _('Spawning'),
    'stoping': _('Stoping'),
    'suspending': _('Suspending'),
    'unpausing': _('Unpausing'),
    'unrescuing': _('Unrescuing'),
    'unshelving': _('Unshelving'),
    'updating_password': _('Updating password'),
}
