import hashlib
import logging

import requests
from OpenSSL.crypto import load_certificate, FILETYPE_PEM

from tionix_vdi_client import exceptions

logger = logging.getLogger(__name__)


def request_token(identity_url, certpath, keypath):
    with open(certpath) as fh:
        cert_data = fh.read()
    cert = load_certificate(FILETYPE_PEM, cert_data)
    subject = cert.get_subject()
    domain_name = dict(subject.get_components()).get('OU')
    issuer_dn = _parse_issuer_dn(cert)
    identity_provider = _get_identity_provider(issuer_dn)
    base_urls = identity_url.split(',')
    errors = []
    for base_url in base_urls:
        url = (
            '{}/v3/OS-FEDERATION/identity_providers/'
            '{}/protocols/x509/auth').format(
                base_url, identity_provider)
        headers = {'X-Domain-Name': domain_name}
        try:
            response = requests.get(
                url, cert=(certpath, keypath), verify=False, headers=headers)
        except requests.RequestException as err:
            logger.error('Token request failed: %s', err)
            errors.append(base_url)
            continue
        if not response.ok:
            logger.error('Token request failed: %s', response.text)
            errors.append(base_url)
            continue
        return response.headers.get('X-Subject-Token')

    raise exceptions.ApplicationConnectionError(
        'Failed to authenticate on: {}'.format(', '.join(errors)))


def _parse_issuer_dn(cert):
    issuer = cert.get_issuer()
    issuer_dn = 'CN={},OU={},O={},ST={},C={}'.format(
        issuer.CN, issuer.OU, issuer.O, issuer.ST, issuer.C)
    return issuer_dn


def _get_identity_provider(issuer_dn):
    return hashlib.sha256(issuer_dn).hexdigest()
