# coding: utf-8

import argparse
import logging
import sys
from sys import platform
from functools import partial

from tionix_vdi_client import MODULE_VERSION
from tionix_vdi_client import app_cli
from tionix_vdi_client.diagnostics import run_diagnostics
from tionix_vdi_client.settings import CONF, SESSION

logger = logging.getLogger('tionix_vdi_client')


# noinspection PyTypeChecker
def create_parser():
    utf8 = partial(str, encoding='utf-8')
    parser = argparse.ArgumentParser(
        description='TIONIX.VDIclient utility',
        conflict_handler='resolve')
    parser.add_argument(
        '--version', action='version', version=MODULE_VERSION)
    parser.add_argument(
        '--username',
        required=False,
        type=utf8,
        help='User name for connection',
    )
    parser.add_argument(
        '--password',
        required=False,
        type=utf8,
        help='Password for connection',
    )
    parser.add_argument(
        '--project',
        required=False,
        type=utf8,
        help='Project for instance retrieval',
    )
    parser.add_argument(
        '--domain',
        required=False,
        type=utf8,
        help='Project domain name',
    )
    parser.add_argument(
        '--server-uri',
        required=False,
        type=utf8,
        help='URI for instance retrieval',
    )
    parser.add_argument(
        '--connect',
        required=False,
        type=utf8,
        help='RDP-client connection command template. '
             'Example: xfreerdp /u:{user} /p:{password}  /v:{ip}',
    )
    parser.add_argument(
        '--no-ui',
        action='store_true',
        required=False,
        help='Provide this flag if there is no need to launch GUI',
    )
    parser.add_argument(
        '--store-session',
        action='store_true',
        required=False,
        help='Provide this flag if there is a need to save user, password, '
             'project, domain to session file',
    )
    parser.add_argument(
        '--diagnostic',
        action='store_true',
        help='Run self-diagnostic'
    )
    parser.add_argument(
        '--debug',
        dest='debug',
        action='store_true',
        help='Activate the debug mode for self-diagnostic'
    )
    parser.add_argument(
        '--ikecfg',
        required=False,
        type=str,
        help='Client IP for the use of the IPSec/IKE',
    )
    return parser


def main():
    args = create_parser().parse_args()

    if args.store_session and not CONF.store_session:
        CONF.store_session = args.store_session
        SESSION.read()

    if args.username:
        if args.username != SESSION.user:
            SESSION.project = ''
            SESSION.password = ''
        SESSION.user = args.username

    if args.password:
        SESSION.password = args.password
    if args.project:
        SESSION.project = args.project

    if args.domain:
        CONF.domain_name = args.domain
    if args.server_uri:
        CONF.cloud = args.server_uri
    if args.connect:
        CONF.connect = args.connect
    if args.ikecfg:
        CONF.ikecfg = args.ikecfg

    if args.diagnostic:
        return run_diagnostics(debug=args.debug)

    if args.no_ui and platform != "win32":
        return app_cli.main()

    # avoid toplevel import to not fall in cli mode
    from tionix_vdi_client import app_gui
    return app_gui.main()


if __name__ == '__main__':
    sys.exit(main())
