# coding: utf-8
import binascii
import codecs
import logging
import os
import subprocess
import sys
import tempfile

from tionix_vdi_client import exceptions
from tionix_vdi_client.i18n import ugettext as _
from tionix_vdi_client.settings import RDP_FILE_SAMPLE_PATH, CONF

logger = logging.getLogger(__name__)

# order is important
RDP_FILE_READ_ENCODINGS = ('utf-8', 'utf-16le')


def read_rdp_config(filepath):
    rdp_file_content = []

    for encoding in RDP_FILE_READ_ENCODINGS:
        try:
            with codecs.open(filepath, encoding=encoding) as fh:
                rdp_file_content = fh.readlines()
        except ValueError:
            continue
        except IOError:
            logger.error("Couldn't read the rdp connect sample config file.")
        else:
            break
    return rdp_file_content


def save_rdp_config(filepath, content):
    with codecs.open(filepath, "w", encoding='utf-16le') as fh:
        for line in content:
            if not line.endswith('\r\n'):
                line += '\r\n'
            fh.write(line)


def _parse_config(content):
    use_multimon = False
    devices = []
    drives = []
    other_devices = False
    domain = ''

    # TODO: move RDP settings read and update to RDPConfig class
    for line in content:
        if line.count(':') >= 2:
            key, _sep, line = line.partition(':')
            _type, _sep, value = line.partition(':')
            value = value.strip().rstrip(';')
            if key == 'domain' and value:
                domain = value
            elif key == 'use multimon' and value == '1':
                use_multimon = True
            elif key == 'usbdevicestoredirect':
                devices = [v.strip() for v in value.split(';')]
            elif key == 'devicestoredirect':
                if value == '*':
                    other_devices = True
            elif key == 'drivestoredirect':
                for drive_name in value.split(';'):
                    drive_name = drive_name.strip()
                    if drive_name:
                        drives.append(drive_name)
    return dict(
        domain=domain,
        use_multimon=use_multimon,
        devices=devices,
        other_devices=other_devices,
        drives=drives,
    )


def _generate_win_rdp_config(host, domain, user, password, ignore_domain):
    from tionix_vdi_client.win.winapi import crypt_protect_data

    rdp_file_content = read_rdp_config(RDP_FILE_SAMPLE_PATH)
    conf = _parse_config(rdp_file_content)
    if ignore_domain or conf['domain']:
        user_name = user
    else:
        user_name = '{0}\\{1}'.format(domain, user)

    password_encrypted = binascii.hexlify(crypt_protect_data(password)).upper()
    rdp_file_content.extend([
        "full address:s:{host}\r\n".format(host=host),
        "UserName:s:{user_name}\r\n".format(user_name=user_name),
        "password 51:b:{password_encrypted}\r\n".format(
            password_encrypted=password_encrypted)
    ])

    rdp_config_path = os.path.join(tempfile.gettempdir(), 'connect.rdp')
    save_rdp_config(rdp_config_path, rdp_file_content)

    return rdp_config_path


def _escape_parameters(parameters, keys_to_escape):
    """
    Escapes specified values in parameters, depending on the OS.
    :param parameters: dict
    :param keys_to_escape: tuple/list
    :return: shallow copy of parameters with modified keys
    """

    def escape_linux(value):
        return "'%s'" % value.replace("'", "'\"'\"'")

    def escape_windows(value):
        return '"%s"' % value

    if sys.platform == "win32":
        escaped = escape_windows
    else:
        escaped = escape_linux

    esc_params = parameters.copy()
    for key in keys_to_escape:
        esc_params[key] = escaped(esc_params[key])

    return esc_params


def rdp_connect(user, password, domain_name, vm_ip, rdp_token,
                stop_event=None, silent=False):
    conn_cmd_params = {
        'ip': vm_ip,
        'user': user,
        'password': password,
        'vdi_token': rdp_token,
        'domain': domain_name
    }

    logger.info("Running connection: user=%s, ip=%s", user, vm_ip)

    if sys.platform == "win32":
        # on Windows platform some special characters in password can't be
        # escaped correctly, so we generate RDP-file with connection
        # parameters to pass to rdp client
        conn_cmd_params['rdp_config'] = _generate_win_rdp_config(
            host=conn_cmd_params['ip'],
            domain=conn_cmd_params['domain'],
            user=conn_cmd_params['user'],
            password=conn_cmd_params['password'],
            ignore_domain=CONF.ignore_domain
        )

    cmd_template = CONF.connect
    bad_cmd_str = _("Invalid parameter: '{connect}'")
    connect_keys = (
        'ip', 'user', 'password', 'project',
        'domain', 'rdp_config', 'vdi_token')
    keys_to_escape = ('user', 'password', 'domain')
    escaped_params = _escape_parameters(
        {key: conn_cmd_params.get(key, '') for key in connect_keys},
        keys_to_escape)
    try:
        cmd = cmd_template.format(**escaped_params)
    except KeyError:
        msg = "Bad command line format %s" % cmd_template
        logger.critical(msg)
        raise exceptions.ApplicationLogicException(bad_cmd_str)

    escaped_params.update({'password': '***'})
    logger.debug('Running command: %s',
                 cmd_template.format(**escaped_params))

    p = subprocess.Popen(
        cmd.encode().decode(sys.getfilesystemencoding()),
        shell=True, stdin=subprocess.PIPE, stdout=subprocess.PIPE,
        stderr=subprocess.PIPE
    )
    if stop_event:
        while p.poll() is None:
            stop_event.wait(0.5)
            if stop_event.is_set():
                p.kill()
                logger.debug('RDP connection canceled')
                return

    out, err = p.communicate()
    if p.returncode:
        message_err = err.encode().decode(sys.stdout.encoding or 'utf-8').strip()
        message_out = out.encode().decode(sys.stdout.encoding or 'utf-8').strip()
        message = message_err or message_out
        logger.error("Remote Access Client sent a message: %s", message)
        if not silent:
            msg = _("Remote Access Client sent a message: {}").format(message)
            err = exceptions.ApplicationLogicException(msg)
            err.exit_code = p.returncode
            raise err
