# coding: utf-8
# pylint: disable=no-name-in-module

import configparser
import base64
import binascii
import codecs
import getpass
import locale
import logging
import logging.config
import os
import shutil
import sys
import tempfile

from tionix_vdi_client import BASE_DIR as ORIGINAL_BASE_DIR, Localization

FS_ENCODING = sys.getfilesystemencoding()
APP_DIR_NAME = '.tionix-vdi-client'
APP_DIR_PATH = os.path.join(os.path.expanduser('~'), APP_DIR_NAME).encode().decode(
    FS_ENCODING)
RDP_FILE_SAMPLE_NAME = 'rdp.conf.sample'
RDP_FILE_SAMPLE_PATH = os.path.join(APP_DIR_PATH, RDP_FILE_SAMPLE_NAME)

if getattr(sys, 'frozen', False):
    # The application is frozen
    BASE_DIR = os.path.dirname(sys.executable)
else:
    BASE_DIR = ORIGINAL_BASE_DIR

BASE_DIR = BASE_DIR.encode().decode(FS_ENCODING)

_settings = None  # Settings cache if reread settings is not required

MAX_LENGTH_CONFIG_MSG = 250

LOG_FILE_NAME = 'tionix-vdi-client'
LOG_FILE_EXT = 'log'

LOG_LEVELS = (
    'CRITICAL',
    'ERROR',
    'WARNING',
    'INFO',
    'DEBUG',
)

LOGGING = {
    'version': 1,
    'disable_existing_loggers': True,
    'formatters': {
        'default': {
            'format': '%(asctime)s.%(msecs)03d %(process)d %(threadName)s '
                      '%(levelname)s %(name)s %(filename)s:%(lineno)d '
                      '[-] %(message)s',
            'datefmt': '%Y-%m-%d %H:%M:%S',
        }
    },
    'handlers': {
        'null': {
            'class': 'logging.NullHandler',
        },
        'console': {
            'level': 'INFO',
            'class': 'tionix_vdi_client.log.BufferingStreamHandler',
            'formatter': 'default',
            'filters': ['token_filter'],
            'buffer_size': 10,  # records
        },
    },
    'loggers': {
        '': {
            'level': 'INFO',
            'handlers': ['console']
        },
        'tionix_vdi_client': {
            'level': 'INFO',
            'handlers': ['console'],
            'propagate': False,
        },
        'urllib3.connectionpool': {
            'level': 'ERROR'
        }
    },
    'filters': {
        'token_filter': {
            '()': 'tionix_vdi_client.log.RedactingFilter',
            'exprs': ["u'vdi_token': u'.*'",
                      "'vdi_token': '.*'",
                      '(vdi_token=.*?|$)\&',
                      '/t:{vdi_token}', '/t:(.*)',
                      '{vdi_token}',
                      'vdi_token',
                      "u'rdp_token': u'.*'",
                      "'rdp_token': '.*'",
                      '(rdp_token=.*?|$)\&',
                      '/t:{rdp_token}',
                      '/t:(.*)',
                      '{rdp_token}',
                      'rdp_token',
                      ("'password': u'.*'", "'password': u'*****'"),
                      ("'password': '.*'", "'password': '*****'"),
                      ("'new_password': u'.*'", "'new_password': u'*****'"),
                      ("'new_password': '.*'", "'new_password': '*****'")]
        },
    },
}

logging.config.dictConfig(LOGGING)

logger = logging.getLogger(__name__)


def get_default_connect_param():
    if sys.platform == "win32":
        conn_cmd = 'mstsc {rdp_config}'
    else:
        conn_cmd = 'xfreerdp /u:{user} /p:{password} /v:{ip} /cert:tofu'
    return conn_cmd


def get_default_locale():
    code, _encoding = locale.getdefaultlocale()
    return Localization.get_lang_by_code(code)


def find_pycs11_lib():
    result = ''
    filename = 'libisbc_pkcs11_main.so'
    paths = (
        os.path.join(os.getcwd().encode().decode(FS_ENCODING), filename),
        os.path.join(APP_DIR_PATH, filename)
    )
    for path in paths:
        if os.path.exists(path):
            result = path
            break
    return result


# noinspection PyTypeChecker
def validate_log_level(v, def_v):
    if isinstance(v, str):
        v = v.upper()
    return def_v if isinstance(logging.getLevelName(v), str) else v


# noinspection PyTypeChecker
def validate_bool(val, default_value=False):
    if isinstance(val, str):
        val = val.lower()
    if val in (True, 'true', 'yes', 'y', '1', 1):
        value = True
    elif val in (False, 'false', 'no', 'n', '0', 0):
        value = False
    else:
        value = default_value
    return value


def validate_use_smartcard(val, default_value):
    val = validate_bool(val, default_value)
    if sys.platform == "win32":
        # smartcard is disabled for Windows
        val = False
    return val


# noinspection PyTypeChecker
def validate_int(val, default_value):
    if isinstance(val, str):
        try:
            val = int(val)
        except ValueError:
            val = default_value
    elif not isinstance(val, int):
        val = default_value
    return val


def validate_positive_int(val, default_value):
    val = validate_int(val, default_value)
    if val < 0:
        val = default_value
    return val


# noinspection PyTypeChecker
def validate_list(val, default_value):
    if not isinstance(val, (list, tuple)):
        if not isinstance(val, str):
            val = default_value
        val = val.split(',')
        val = [v.strip() for v in val if v.strip()]
    return val


def validate_str(val, def_v=''):
    val = val.strip()
    if not val:
        val = def_v
    return val


# pylint: disable-msg=W0702
# noinspection PyBroadException
def validate_max_length(val, default):
    if not val:
        return default
    elif len(val) > MAX_LENGTH_CONFIG_MSG:
        logger.debug("The allowed message length has been exceed 250 "
                     "characters. The message will be cut off.")
    return val[:MAX_LENGTH_CONFIG_MSG]


def validate_log_file_location(val, def_value):
    result = def_value
    if val:
        val = os.path.expandvars(val)
        if not os.path.exists(val):
            os.makedirs(val)

        if os.path.isdir(val):
            result = val
    return result


def validate_language(val, def_value):
    result = def_value
    available_languages = list(Localization.locales_dict.keys())
    # noinspection PyTypeChecker
    if isinstance(val, str):
        val = val.lower().strip()
        if val in available_languages:
            result = val
        else:
            logger.warning(
                'Language "%s" is not supported. Available languages are:'
                ' "%s". Your system language is "%s", so it is set as the '
                'default language.',
                val, ', '.join(available_languages), def_value)
    return result


def validate_file(val, def_value):
    value = os.path.expandvars(val)
    if not os.path.exists(value) and def_value:
        value = def_value
    return value


def validate_oid(val, def_value):
    if not all(n.isdigit() for n in val.split('.')):
        val = def_value
    return val


def get_log_file_path(dir_path=None):
    dir_path = dir_path if dir_path else APP_DIR_PATH
    dir_path = os.path.abspath(dir_path)

    current_user = getpass.getuser().encode().decode(FS_ENCODING)

    available_names = (
        '{0}.{1}'.format(LOG_FILE_NAME, LOG_FILE_EXT),
        '{0}-{1}.{2}'.format(LOG_FILE_NAME, current_user, LOG_FILE_EXT),
    )

    for name in available_names:
        if os.access(dir_path, os.W_OK):
            path = os.path.join(dir_path, name)
            if not os.path.exists(path) or os.access(path, os.W_OK):
                break
    else:
        _id, path = tempfile.mkstemp(
            prefix=LOG_FILE_NAME + '-', suffix='.' + LOG_FILE_EXT)
        path.encode().decode(FS_ENCODING)
    return path


def update_nested_dict(initial_dict, update_dict):
    for key, val in list(update_dict.items()):
        if isinstance(val, dict):
            initial_dict[key] = update_nested_dict(
                initial_dict.get(key, {}), val)
        else:
            initial_dict[key] = val
    return initial_dict


class ConfigReadError(Exception):
    pass


class UnicodeConfigParser(configparser.ConfigParser):
    FILE_READ_ENCODINGS = ('utf-8-sig', 'cp1251')

    def write(self, fp):
        """Write an .ini format configuration file

        Overridden `ConfigParser.write` method to support unicode values
        and platform-dependent newlines.
        """
        linesep = os.linesep
        if self._defaults:
            fp.write("[{section}]".format(section=configparser.DEFAULTSECT))
            fp.write(linesep)
            for (key, value) in list(self._defaults.items()):
                if isinstance(value, str):
                    value = value.encode().decode('utf-8')
                value = str(value).replace('\n', '\n\t')
                fp.write("{key} = {value}".format(key=key, value=value))
                fp.write(linesep)
            fp.write(linesep)
        for section in self._sections:
            fp.write("[{section}]".format(section=section))
            fp.write(linesep)
            for (key, value) in list(self._sections[section].items()):
                if key == "__name__":
                    continue
                if (value is not None) or (self._optcre == self.OPTCRE):
                    if isinstance(value, str):
                        value = value.encode().decode('utf-8')
                    value = str(value).replace('\n', '\n\t')
                    fp.write("{key} = {value}".format(key=key, value=value))
                    fp.write(linesep)
            fp.write(linesep)

    def safe_get(self, *args, **kwargs):
        val = kwargs.pop('default', None)
        try:
            val = self.get(*args, **kwargs)
        except configparser.NoOptionError:
            pass
        return val

    def read_file(self, filepath):
        """Read from configuration file

        All values are in UTF-8 encoding.
        """
        for encoding in self.FILE_READ_ENCODINGS:
            try:
                self.read(filepath, encoding=encoding)
            except ValueError:
                continue
            except IOError as exc:
                logger.error(
                    'Error while reading file %s: %s', filepath, exc)
                raise ConfigReadError()
            except (configparser.MissingSectionHeaderError,
                    configparser.ParsingError) as exc:
                logger.error(
                    'Error while reading file %s: %s', filepath, exc)
                raise ConfigReadError()
            else:
                break
        else:
            logger.error('Failed to detect encoding of file %s', filepath)
            raise ConfigReadError()

    def write_file(self, filepath):
        """Save config  to .ini format file with UTF-8 encoding"""

        with codecs.open(filepath, 'w', 'utf-8') as fp:
            self.write(fp)


class Settings(dict):
    DOMAIN_NAME = 'domain_name'
    CLOUD = 'cloud'
    IGNORE_DOMAIN = 'ignore_domain'
    SECONDARY_CLOUD = 'secondary_cloud'
    CONNECT = 'connect'
    LANGUAGE = 'language'
    required_attributes = ()

    _attributes = [DOMAIN_NAME, CLOUD, IGNORE_DOMAIN, SECONDARY_CLOUD, CONNECT,
                   LANGUAGE]
    _extra_attributes = []

    _config_class = UnicodeConfigParser

    def __getattr__(self, name):
        if name in self._attributes or name in self._extra_attributes:
            return self.get(name, '')
        else:
            msg = "'{0}' object has no attribute '{1}'".format(
                type(self).__name__, name)
            raise AttributeError(msg)

    def __setattr__(self, item, value):
        if item in self._attributes or item in self._extra_attributes:
            self[item] = value
        else:
            msg = "'{0}' object has no attribute '{1}'".format(
                type(self).__name__, item)
            raise AttributeError(msg)

    def load_config(self, filepath):
        config = self._config_class()
        config.read_file(filepath)
        return config


class SessionSettings(Settings):
    USER = 'user'
    PASSWORD = 'password'
    PROJECT = 'project'
    required_attributes = (USER, PASSWORD, Settings.DOMAIN_NAME)

    _attributes = Settings._attributes + [USER, PASSWORD, PROJECT]
    _extra_attributes = [
        'ip', 'token', 'rdp_token', 'smartcard_used', 'smartcard_session',
        'token_auth']
    _encode_keys = (USER, PASSWORD, PROJECT, Settings.DOMAIN_NAME)
    _session_section_name = 'DEFAULT'

    @property
    def _session_file_name(self):
        return os.path.join(APP_DIR_PATH, 'tnx_client_session')

    def read(self):
        resave = False
        try:
            config = self.load_config(self._session_file_name)
        except ConfigReadError:
            pass
        else:
            for key, value in config.items(self._session_section_name):
                if key == self.PASSWORD and not CONF.store_password and value:
                    resave = True
                    value = ""
                elif key in self._encode_keys:
                    try:
                        value = base64.decodebytes(value.encode()).decode('utf-8')
                    except binascii.Error:
                        logger.error("Failed to decode {}. Value is "
                                     "considered as empty.".format(key))
                        value = ""
                setattr(self, key, value)
        if resave:
            self.save()

    def save(self):
        config = self._config_class()
        for key, value in list(self.items()):
            if key not in self._attributes:
                continue
            if value is None:
                value = ''
            if key in self._encode_keys:
                if not isinstance(value, str):
                    value = value.encode().decode('utf-8')
                value = base64.encodebytes(value.encode()).decode('utf-8')
            if key == self.PASSWORD and not CONF.store_password:
                value = ''
            config.set(self._session_section_name, key, value)
        try:
            config.write_file(self._session_file_name)
        except IOError as exc:
            logger.warning('Error while saving session to '
                           'file: %s', exc)

    def clear(self):
        for attr in self._attributes:
            setattr(self, attr, '')


class ConfigSettings(Settings):
    LOG_LEVEL = 'log_level'
    STORE_SESSION = 'store_session'
    STORE_PASSWORD = 'store_password'
    LOG_FILE_LOCATION = 'log_file_location'
    PROJECT = 'project'
    SHOW_SETTINGS = 'show_settings'
    PYKCS11_LIB = 'PYKCS11LIB'
    IKECFG = 'ikecfg'
    ADDITIONAL_CLOUDS = 'additional_clouds'
    RETRIES = 'retries'
    TIMEOUT = 'timeout'
    SMART_OID = 'SMARTCARD_OID'
    USE_SMARTCARD = 'use_smartcard'
    SUPPORT_MESSAGE_RU = 'contact_support_message_ru'
    SUPPORT_MESSAGE_EN = 'contact_support_message_en'
    PASSWORD_GENERATION = 'password_generation'
    GET_VM_TIMEOUT = 'get_vm_timeout'
    WEB_GUARD = 'web_guard'
    USE_CERT = 'use_cert'
    CERT = 'cert'
    KEY = 'key'
    IDENTITY_URL = 'identity_url'
    MINIMIZE_TO_TRAY = 'minimize_to_tray'
    SINGLE_LAUNCH = 'single_launch'
    SILENT = 'silent'

    required_attributes = (Settings.CLOUD, Settings.CONNECT)

    _attributes = Settings._attributes + [
        LOG_LEVEL, STORE_SESSION, STORE_PASSWORD, LOG_FILE_LOCATION, PROJECT,
        SHOW_SETTINGS, PYKCS11_LIB, IKECFG, ADDITIONAL_CLOUDS,
        RETRIES, TIMEOUT, SMART_OID, USE_SMARTCARD,
        SUPPORT_MESSAGE_RU, SUPPORT_MESSAGE_EN, PASSWORD_GENERATION,
        GET_VM_TIMEOUT, WEB_GUARD, USE_CERT, CERT, KEY, IDENTITY_URL,
        MINIMIZE_TO_TRAY, SINGLE_LAUNCH, SILENT
    ]

    _config_section_name = 'DEFAULT'
    _conf_file = 'client.conf'
    _sample_file = 'client.conf.sample'

    ordered_file_variables = (
        Settings.CLOUD,
        Settings.IGNORE_DOMAIN,
        Settings.SECONDARY_CLOUD,
        Settings.DOMAIN_NAME,
        LOG_LEVEL,
        Settings.CONNECT,
        LOG_FILE_LOCATION,
        STORE_SESSION,
        STORE_PASSWORD,
        Settings.LANGUAGE,
        PROJECT,
        SHOW_SETTINGS,
        PYKCS11_LIB,
        SMART_OID,
        IKECFG,
        ADDITIONAL_CLOUDS,
        RETRIES,
        TIMEOUT,
        USE_SMARTCARD,
        SUPPORT_MESSAGE_RU,
        SUPPORT_MESSAGE_EN,
        PASSWORD_GENERATION,
        GET_VM_TIMEOUT,
        WEB_GUARD,
        USE_CERT,
        CERT,
        KEY,
        IDENTITY_URL,
        MINIMIZE_TO_TRAY,
        SINGLE_LAUNCH,
        SILENT,
    )

    default = {
        Settings.CLOUD: '',
        Settings.IGNORE_DOMAIN: False,
        Settings.SECONDARY_CLOUD: '',
        Settings.CONNECT: get_default_connect_param(),
        Settings.DOMAIN_NAME: 'default',
        LOG_LEVEL: 'INFO',
        STORE_SESSION: True,
        STORE_PASSWORD: False,
        LOG_FILE_LOCATION: '',
        Settings.LANGUAGE: get_default_locale(),
        PROJECT: True,
        SHOW_SETTINGS: True,
        PYKCS11_LIB: find_pycs11_lib(),
        SMART_OID: '',
        IKECFG: '',
        ADDITIONAL_CLOUDS: '',
        RETRIES: 2,
        TIMEOUT: 15,
        USE_SMARTCARD: True,
        SUPPORT_MESSAGE_RU: 'Обратитесь к системному администратору.',
        SUPPORT_MESSAGE_EN: 'Please contact system administrator.',
        PASSWORD_GENERATION: False,
        GET_VM_TIMEOUT: 5,
        WEB_GUARD: False,
        USE_CERT: False,
        CERT: '',
        KEY: '',
        IDENTITY_URL: '',
        MINIMIZE_TO_TRAY: False,
        SINGLE_LAUNCH: False,
        SILENT: False,
    }

    # values that will not be written in config unless explicitly set
    hidden_defaults = {
        USE_SMARTCARD,
        PASSWORD_GENERATION,
        GET_VM_TIMEOUT,
    }

    _validators = {
        Settings.CONNECT: validate_str,
        Settings.IGNORE_DOMAIN: validate_bool,
        LOG_LEVEL: validate_log_level,
        STORE_SESSION: validate_bool,
        STORE_PASSWORD: validate_bool,
        LOG_FILE_LOCATION: validate_log_file_location,
        Settings.LANGUAGE: validate_language,
        PROJECT: validate_bool,
        SHOW_SETTINGS: validate_bool,
        PYKCS11_LIB: validate_file,
        SMART_OID: validate_oid,
        ADDITIONAL_CLOUDS: validate_list,
        RETRIES: validate_int,
        TIMEOUT: validate_int,
        USE_SMARTCARD: validate_use_smartcard,
        SUPPORT_MESSAGE_RU: validate_max_length,
        SUPPORT_MESSAGE_EN: validate_max_length,
        PASSWORD_GENERATION: validate_bool,
        GET_VM_TIMEOUT: validate_positive_int,
        WEB_GUARD: validate_bool,
        USE_CERT: validate_bool,
        CERT: validate_file,
        KEY: validate_file,
        MINIMIZE_TO_TRAY: validate_bool,
        SINGLE_LAUNCH: validate_bool,
        SILENT: validate_bool,
    }

    @property
    def _sample_config_name(self):
        return os.path.join(BASE_DIR, self._sample_file)

    @property
    def _default_config_name(self):
        return os.path.join(APP_DIR_PATH, self._conf_file)

    def __init__(self):
        super(ConfigSettings, self).__init__()
        self.create_configuration_file()

    def create_configuration_file(self):
        if not os.path.isfile(self._default_config_name):
            try:
                shutil.copy(self._sample_config_name,
                            self._default_config_name)
                self.save()
            except IOError:
                logger.info("Couldn't create config file, using defaults!")

    @classmethod
    def validator(cls, setting_param):
        return cls._validators.get(setting_param, lambda x, def_x: x)

    def _read_config(self):
        try:
            config = self.load_config(self._default_config_name)
        except ConfigReadError:
            result = self.default
        else:
            result = {}
            for setting_param in self.ordered_file_variables:
                validator = self.validator(setting_param)
                default = self.default[setting_param]
                result[setting_param] = validator(
                    config.safe_get(self._config_section_name, setting_param,
                                    default=default), default)
        return result

    def read(self, reread=False):
        """
        Read the settings file
        """
        if not reread and self:
            return
        self.update(self.default)
        self.update(self._read_config())
        update_logging_level()

    def save(self):
        """
        Save settings to file
        """
        config = self._config_class()
        if list(self.items()):
            for key in self.ordered_file_variables:
                value = getattr(self, key)
                if key in self.hidden_defaults and value == self.default[key]:
                    continue
                if isinstance(value, (list, tuple)):
                    value = ','.join(value)
                config.set(self._config_section_name, key, str(value))
        try:
            config.write_file(self._default_config_name)
        except IOError as exc:
            logger.warning('Error while saving settings to '
                           'file: %s', exc)


def update_logging_level():
    """Update logging level from config."""
    config = CONF
    logging.config.dictConfig(get_logging_config(
        level=config[config.LOG_LEVEL],
        log_file_dir_path=config[config.LOG_FILE_LOCATION]))


# noinspection PyTypeChecker
def get_logging_config(level=None, log_file_dir_path=None):
    level = level or ConfigSettings.default[ConfigSettings.LOG_LEVEL]

    # Have to update log level before updating other variables for
    # proper logger output
    logger.setLevel(level)

    log_file_dir_path = log_file_dir_path or ConfigSettings.default[
        ConfigSettings.LOG_FILE_LOCATION]

    previous_log_path = LOGGING.get('handlers', {}).get(
        'file', {}).get('filename')

    if (previous_log_path and os.path.abspath(
            os.path.dirname(previous_log_path)) == os.path.abspath(
                log_file_dir_path)):
        log_file_path = previous_log_path
    else:
        log_file_path = get_log_file_path(log_file_dir_path)

    if previous_log_path:
        if previous_log_path != log_file_path:
            logger.info('Log file location changed from "%s" to "%s"',
                        previous_log_path, log_file_path)
    else:
        logger.info('Log file path: %s', log_file_path)

    custom_config = {
        'handlers': {
            'null': {
                'level': level,
            },
            'console': {
                'level': level,
            },
            'file': {
                'class': 'logging.FileHandler',
                'filename': log_file_path,
                'formatter': 'default',
                'encoding': 'utf-8',
            }
        },
        'loggers': {
            '': {
                'level': level,
                'handlers': ['console', 'file']
            },
            'tionix_vdi_client': {
                'level': level,
                'handlers': ['console', 'file'],
                'propagate': False
            },
        }
    }
    update_nested_dict(LOGGING, custom_config)
    return LOGGING


# Create directory for application files
if not os.path.exists(APP_DIR_PATH):
    os.mkdir(APP_DIR_PATH)

CONF = ConfigSettings()
CONF.read(reread=True)

SESSION = SessionSettings()
if CONF.store_session:
    SESSION.read()

if sys.platform == "win32":
    if not os.path.isfile(RDP_FILE_SAMPLE_PATH):
        try:
            shutil.copy(os.path.join(BASE_DIR, RDP_FILE_SAMPLE_NAME),
                        RDP_FILE_SAMPLE_PATH)
        except IOError as err:
            logger.error("Couldn't create rdp connect sample config file."
                         " %s", str(err))
