# coding: utf-8
# pylint: disable=import-error,no-name-in-module

import logging
import os
from collections import namedtuple
from datetime import datetime

from OpenSSL.crypto import FILETYPE_ASN1, load_certificate
from PyKCS11 import (
    CKA_CLASS, CKA_LABEL, CKA_PRIVATE, CKA_TOKEN, CKA_VALUE,
    CKF_RW_SESSION, CKF_SERIAL_SESSION, CKO_CERTIFICATE,
    CKO_DATA, CKR_PIN_INCORRECT, CK_TRUE, PyKCS11Error, PyKCS11Lib
)
from pyasn1.codec.der.decoder import decode
from pyasn1.type.univ import ObjectIdentifier, Sequence, SequenceOf

from tionix_vdi_client.i18n import ugettext as _
from tionix_vdi_client.settings import CONF
from tionix_vdi_client.smart_card import exceptions
from tionix_vdi_client.smart_card.pcsc import PCSCClient, PCSCError

logger = logging.getLogger(__name__)

Certificate = namedtuple(
    'Certificate',
    ['subject_email', 'issuer_email', 'start_date', 'end_date', 'oids'])

SMARTCARD_ERROR_MSG = (
    "Couldn't connect to smart card. "
    "Login/Pass will be searched in config files. "
    "Error: %s")


def _retrieve_username(session, oid=None):
    user_certificate = _find_user_certificate(session, oid)
    user_email = user_certificate.subject_email
    return user_email.split('@')[0].encode().decode('utf-8')


def _get_objects(session, template):
    try:
        vdi_files = session.findObjects(template)
    except PyKCS11Error as err:
        logger.warning('Error while searching files on smart card: %s.', err)
        vdi_files = []
    return vdi_files


def _delete_object(session, obj):
    try:
        session.destroyObject(obj)
    except PyKCS11Error as err:
        logger.warning('Error while deleting file on smart card: %s', err)


def _get_password_template(login):
    if isinstance(login, str):
        login = login.encode().decode('utf-8')
    template = [
        (CKA_CLASS, CKO_DATA),
        (CKA_TOKEN, CK_TRUE),
        (CKA_LABEL, login),
    ]
    return template


def _create_object(session, template):
    try:
        result = session.createObject(template)
    except PyKCS11Error as err:
        logger.warning('Error while creating file on smart card: %s.', err)
        result = None
    return result


def _get_oids_from_x509cert(x509_cert):
    oids = []
    try:
        for i in range(0, x509_cert.get_extension_count()):
            ext = x509_cert.get_extension(i)
            if ext.get_short_name() != 'extendedKeyUsage':
                continue
            eku = ext.get_data()
            for eku_val in decode(eku):
                if not isinstance(eku_val, (Sequence, SequenceOf)):
                    continue
                # noinspection PyProtectedMember
                for v in eku_val._componentValues:
                    if isinstance(v, ObjectIdentifier):
                        oids.append(v.prettyPrint())
    except Exception as err:
        logger.warning('Failed to parse certificate OIDs: %s', err)
    return oids


def _get_certificates(session):
    certificates = _get_objects(session, [(CKA_CLASS, CKO_CERTIFICATE)])

    parsed_certificates = []
    extracting_attrs = [CKA_VALUE]
    for cert in certificates:
        # noinspection PyBroadException
        try:
            attributes = session.getAttributeValue(cert, extracting_attrs)
            attributes = dict(list(zip(extracting_attrs, attributes)))
            x509_cert = load_certificate(
                FILETYPE_ASN1, str(bytearray(attributes[CKA_VALUE])))
        except Exception as err:
            logger.debug('Failed to load certificate: %s', err)
            continue

        subject_email = x509_cert.get_subject().emailAddress or ''
        issuer_email = x509_cert.get_issuer().emailAddress or ''
        end_date = datetime.strptime(
            x509_cert.get_notAfter(), '%Y%m%d%H%M%SZ')
        start_date = datetime.strptime(
            x509_cert.get_notBefore(), '%Y%m%d%H%M%SZ')
        oids = _get_oids_from_x509cert(x509_cert)
        parsed_certificates.append(Certificate(
            subject_email=subject_email,
            issuer_email=issuer_email,
            start_date=start_date,
            end_date=end_date,
            oids=oids))

    return sorted(
        parsed_certificates, key=lambda c: c.start_date, reverse=True)


def _find_user_certificate(session, oid=None):
    certificates = _get_certificates(session)
    if not certificates:
        raise exceptions.CertificateNotFound()
    logger.debug(
        'Found following certificates:\n%s',
        '\n'.join(
            '{{subject_email={0}, issuer_email={1}, '
            'start_date={2}, end_date={3}, OIDs={4}}}'.format(
                c.subject_email, c.issuer_email, c.start_date, c.end_date,
                c.oids) for c in certificates))

    now = datetime.utcnow()
    filtered_certs = []
    for k, cert in enumerate(certificates, start=1):
        logger.debug('Processing certificate #%s', k)

        if not cert.subject_email:
            logger.debug('Certificate does not contain subject email')
            continue

        if cert.subject_email == cert.issuer_email:
            logger.debug("Skipping self-signed certificate")
            continue

        if oid and oid not in cert.oids:
            logger.debug(
                'Certificate does not contain required EKU OID %s', oid)
            continue

        if cert.end_date < now:
            logger.debug('Certificate expired')
            continue

        filtered_certs.append(cert)

    if not filtered_certs:
        raise exceptions.AppropriateCertificateNotFound()
    elif len(filtered_certs) > 1:
        if filtered_certs[0].start_date == filtered_certs[1].start_date:
            raise exceptions.MultipleCertificatesFound()

    return filtered_certs[0]


def _retrieve_password(session, username):
    result_password = None
    template = _get_password_template(username)
    vdi_files = _get_objects(session, template)
    if vdi_files:
        vdi_file = vdi_files[0]
        try:
            attributes = session.getAttributeValue(vdi_file, [CKA_VALUE])
            result_password = str(bytearray(attributes[0])).encode().decode('utf-8')
        except PyKCS11Error as err:
            logger.warning(
                'Error while reading vdi file: %s.', err)
    return result_password


def get_session():
    if CONF.PYKCS11LIB and "PYKCS11LIB" not in os.environ:
        if os.path.exists(CONF.PYKCS11LIB):
            os.environ["PYKCS11LIB"] = CONF.PYKCS11LIB.encode.decode('utf-8')
        else:
            logger.warning(
                "PYKCS11LIB file is not found: %s", CONF.PYKCS11LIB)
            raise exceptions.SmartCardNotConfigured()

    pkcs11 = PyKCS11Lib()

    try:
        pkcs11.load()
    except PyKCS11Error as err:
        logger.warning(
            'Couldn"t load pkcs11 lib: %s. Please check "PYKCS11LIB" option'
            ' in the configuration file', err)
        raise exceptions.SmartCardNotConfigured()

    logger.debug('PKCS11 lib initialized.')

    slots = pkcs11.getSlotList()
    if not slots:
        logger.warning('No card readers found. Login/Pass will be '
                       'searched in config files.')
        raise exceptions.SmartCardNotConfigured()

    slot = slots[0]
    logger.info('Found card reader (slot): %s',
                pkcs11.getSlotInfo(slot).slotDescription)

    try:
        token_info = pkcs11.getTokenInfo(slot)
    except PyKCS11Error as err:
        logger.error('Failed to obtain token info: %s.', err)
        raise exceptions.SmartCardNotConfigured()

    try:
        session = pkcs11.openSession(
            slot, CKF_SERIAL_SESSION | CKF_RW_SESSION)
    except PyKCS11Error as err:
        logger.warning(SMARTCARD_ERROR_MSG, err)
        raise exceptions.SmartCardNotConfigured()

    session.min_pin_len = token_info.ulMinPinLen
    session.max_pin_len = token_info.ulMaxPinLen

    return session


def get_pin_attempts():
    attempts = None
    try:
        with PCSCClient() as pcsc:
            attempts = pcsc.get_pin_attempts()
    except PCSCError as err:
        logger.warning('Failed to get pin attempts: %s', err)
    return attempts


def get_session_data(session, pin):
    try:
        session.login(pin)
    except PyKCS11Error as err:
        if err.value == CKR_PIN_INCORRECT:
            error_msg = _("You've entered an incorrect PIN.")
            logger.error(error_msg)
            raise exceptions.IncorrectPin(error_msg)

        logger.warning(SMARTCARD_ERROR_MSG, err)
        raise exceptions.SmartCardNotConfigured()

    username = _retrieve_username(session, oid=CONF.SMARTCARD_OID)
    password = _retrieve_password(session, username)
    return username, password


def update_password(session, username, new_password):
    template = _get_password_template(username)
    vdi_files = _get_objects(session, template)
    if vdi_files:
        vdi_file = vdi_files[0]
        _delete_object(session, vdi_file)

    store_template = template + [
        (CKA_VALUE, new_password.encode().decode('utf-8')),
        (CKA_PRIVATE, CK_TRUE)
    ]
    return _create_object(session, store_template)
