# coding: utf-8
"""
PC/SC client for sending APDU commands to smart card.

Esmart APDU commands for PIN verification (user):

    Select user key (path is /3F00/8F01/7F01):
        00 A4 00 00 02 3F 00
        00 A4 01 00 02 8F 01
        00 A4 01 00 02 7F 01

    Verify pin:
        00 20 00 83

Esmart commands for PIN verification (SO):

    Select SO key (path is /3F00/8F01):
        00 A4 00 00 02 3F 00
        00 A4 01 00 02 8F 01

    Verify pin:
        00 20 00 81
"""

import logging

from collections import namedtuple

try:
    from smartcard.System import readers  # pylint: disable=import-error
except ImportError:
    readers = None


logger = logging.getLogger(__name__)

SELECT_P1_ROOT = 0x00
SELECT_P1_CHILD = 0x01
SELECT_SW1_SUCCESS = 0x61


VendorData = namedtuple(
    'VendorData',
    ['user_key', 'so_key', 'user_verify_mode', 'so_verify_mode'])


class CardVendors(object):
    ESMART = 1
    DEFAULT = ESMART

    VENDOR_DATA = {
        ESMART: VendorData(
            user_key=([0x3F, 0x00], [0x8F, 0x01], [0x7F, 0x01]),
            so_key=([0x3F, 0x00], [0x8F, 0x01]),
            user_verify_mode=0x83,
            so_verify_mode=0x81)
    }


class PCSCError(Exception):
    pass


class ApduCommand(object):
    """Base APDU (Application Protocol Data Unit) command.

    Command attributes:
        cla - class of instruction;
        ins - instruction code;
        p1 - instruction parameter 1;
        p2 - instruction parameter 2;
        lc - number of bytes present in the data field of the command;
        data - bytes sent in the data field of the command;
    """

    cla = 0x00  # no secure messaging
    ins = p1 = p2 = lc = data = None

    def __init__(self, data=None, **kwargs):
        if data:
            self.lc = len(data)
            self.data = data

    def to_bytes(self):
        command = [self.cla, self.ins, self.p1, self.p2]
        if self.lc:
            command.append(self.lc)
            command.extend(self.data)
        return command


class SelectFile(ApduCommand):
    """SELECT FILE APDU command

    Select File sets a current file within a logical channel.
    """
    ins = 0xA4
    p2 = 0x00  # select by identifier


class Verify(ApduCommand):
    """VERIFY APDU command

    Initiates the comparison in the card of the verification data sent
    from the interface device with the reference data stored in the card.
    """
    ins = 0x20
    p1 = 0x00
    p2 = 0x81

    # noinspection PyUnusedLocal
    def __init__(self, data=None, **kwargs):
        super(Verify, self).__init__(data=data, **kwargs)
        p2 = kwargs.get('p2')
        if p2 is not None:
            self.p2 = p2


class PCSCClient(object):
    _connection = None
    _auth_path = None
    _verify_p2 = None

    def __init__(self, vendor=CardVendors.DEFAULT):
        self._auth_path = CardVendors.VENDOR_DATA[vendor].user_key
        self._verify_p2 = CardVendors.VENDOR_DATA[vendor].user_verify_mode

    @property
    def connection(self):
        if not self._connection:
            if readers is None:
                raise PCSCError('pyscard package is not installed.')
            # noinspection PyCallingNonCallable
            scard_readers = readers()
            if not scard_readers:
                raise PCSCError('Smart card readers not found.')
            reader = scard_readers[0]
            self._connection = reader.createConnection()
            self._connection.connect()
        return self._connection

    def transmit(self, command):
        command_bytes = command.to_bytes()
        logger.debug('Sending APDU command: {0}'.format(
            ' '.join('{:02X}'.format(i) for i in command_bytes)))
        print()
        data, sw1, sw2 = self.connection.transmit(command_bytes)
        logger.debug(
            'Response status words: {0:02X} {1:02X}'.format(sw1, sw2))
        return data, sw1, sw2

    def select_file(self, file_path):
        for i, identifier in enumerate(file_path):
            select_cmd = SelectFile(data=identifier)
            select_cmd.p1 = SELECT_P1_ROOT if i == 0 else SELECT_P1_CHILD
            _data, sw1, _sw2 = self.transmit(select_cmd)
            if sw1 != SELECT_SW1_SUCCESS:
                raise PCSCError('SELECT FILE failed.')

    def verify_pin(self, pin=None):
        self.select_file(self._auth_path)
        verify_cmd = Verify(data=pin, p2=self._verify_p2)
        _data, sw1, sw2 = self.transmit(verify_cmd)
        success = (sw1, sw2) == (0x90, 0x00)
        return success, sw1, sw2

    def get_pin_attempts(self):
        _success, sw1, sw2 = self.verify_pin()
        if (sw1, sw2) == (0x69, 0x83):
            attempts = 0
        elif sw1 == 0x63 and 0xC1 <= sw2 <= 0xCF:
            attempts = sw2 - 0xC0
        else:
            raise PCSCError('Failed to get PIN attempts.')
        return attempts

    def disconnect(self):
        self.connection.disconnect()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.disconnect()
