# coding: utf-8

import unittest

# noinspection PyPep8Naming
import configparser as configparser

from tionix_vdi_client.i18n import ugettext as _, ungettext
from tionix_vdi_client.settings import ConfigSettings, SessionSettings


class ConfigFilesTest(unittest.TestCase):
    msg_save_failed = _("Failed save to {0} file. "
                        "Please check write permissions of {1}.")
    config_transl = _('config')
    session_transl = _('session')

    @staticmethod
    def find_empty_parameters(settings):
        # type: (ConfigSettings or SessionSettings) -> list
        return [
            key for key in settings.required_attributes if
            key not in settings or settings[key] in {None, ''}
        ]

    def test_config_file_read_save(self):
        """Checks write permissions for config file.

        :lang:ru Проверяет наличие прав на запись в файл конфигурации.
        """
        conf_parser = configparser.ConfigParser()
        conf_settings_location = ConfigSettings()._default_config_name

        try:
            conf_parser.read(conf_settings_location)
        except (configparser.ParsingError,
                configparser.MissingSectionHeaderError):
            self.fail(_("Failed to parse config file."))

        test_cloud = 'test.loc'

        config_file = ConfigSettings()
        config_file.read(reread=True)

        orig_cloud = config_file.cloud
        config_file.cloud = test_cloud
        config_file.save()

        config_file.read(reread=True)

        msg_failed_save_config = self.msg_save_failed.format(
            self.config_transl, conf_settings_location)

        self.assertNotEqual(orig_cloud, config_file.cloud,
                             msg=_(msg_failed_save_config))

        config_file.cloud = orig_cloud
        config_file.save()

        config_file.read(reread=True)

        self.assertEqual(orig_cloud, config_file.cloud,
                          msg=_(msg_failed_save_config))

    def test_session_file_read_save(self):
        """Checks write permissions for session file.

        :lang:ru Проверяет наличие прав на запись в файл сессии.
        """
        conf_parser = configparser.ConfigParser()
        session_file_location = SessionSettings()._session_file_name
        msg_failed_save_session = self.msg_save_failed.format(
            self.session_transl, session_file_location)

        test_username = 'tester'

        sessions_settings = SessionSettings()
        try:
            conf_parser.read(session_file_location)
        except (configparser.ParsingError,
                configparser.MissingSectionHeaderError):
            self.fail(_("Failed to parse session file."))

        sessions_settings.read()
        orig_user = sessions_settings.user
        sessions_settings.user = test_username
        sessions_settings.save()

        sessions_settings.read()

        self.assertEqual(test_username, sessions_settings.user,
                         msg=msg_failed_save_session)

        sessions_settings.user = orig_user
        sessions_settings.save()

        sessions_settings.read()

        self.assertEqual(orig_user, sessions_settings.user,
                         msg=msg_failed_save_session)

    def test_config_file_has_required_values(self):
        """Checks that all required values are present in config file.

        :lang:ru Проверяет наличие всех обязательных параметров в файле
         конфигурации.
        """
        config = ConfigSettings()
        config.read(reread=True)
        missing_params = self.find_empty_parameters(config)
        if missing_params:
            self.fail(ungettext(
                'Parameter "{}" is missing in config '
                'file.',
                'Parameters "{}" are missing in config '
                'file.',
                len(missing_params)).format(", ".join(missing_params)))

    def test_session_file_has_required_values(self):
        """Checks that all required values are present in session file.

        :lang:ru Проверяет наличие всех обязательных параметров в файле сессии.
        """
        session_settings = SessionSettings()
        session_settings.read()
        missing_params = self.find_empty_parameters(session_settings)
        if missing_params:
            self.fail(ungettext(
                'Parameter "{}" is missing in session '
                'file.',
                'Parameters "{}" are missing in session '
                'file.',
                len(missing_params)).format(", ".join(missing_params)))
