# coding: utf-8

import functools
import unittest

import mock
import requests

import tionix_vdi_client.main
import tionix_vdi_client.rdp
import tionix_vdi_client.app_cli
from tionix_vdi_client import CONNECTION_TYPE, exceptions, client
from tionix_vdi_client.helpers import RequestStatus
from tionix_vdi_client.tests.unit import stubs
from tionix_vdi_client.client import VM


class TestVdiClientAPIMethods(unittest.TestCase):
    default_credentials = ('tester', '123', 'test')

    def setUp(self):
        conf = stubs.TempConfig()
        conf.read()
        self.addCleanup(conf.delete_config)
        patcher = mock.patch('tionix_vdi_client.client.CONF', conf)
        self.conf = patcher.start()
        self.addCleanup(patcher.stop)

    def list_callable_vdi_actions(self):
        """Provide list of all VDI actions."""
        vdi_client = client.VdiClient(*self.default_credentials)
        vdi_actions = (
            vdi_client.request_vm,
            functools.partial(vdi_client.get_vm, 'some-request-id'),
            functools.partial(vdi_client.change_password, 'somePassword'),
        )
        return vdi_actions

    @mock.patch('tionix_vdi_client.client.requests.post')
    def test_change_password_success(self, mocked_post):
        """Change password request sent and updated in client."""
        new_password = 'МойНовыйПар0ль'
        response = stubs.Response(json={'msg': 'Password changed.'})
        mocked_post.return_value = response

        vdi_client = client.VdiClient(*self.default_credentials)
        vdi_client.change_password(new_password)

        mocked_post.asser_called_once()
        self.assertEqual(vdi_client.auth_params['password'], new_password)

    @mock.patch('tionix_vdi_client.client.requests.post')
    def test_request_vm_got_request_id(self, mocked_post):
        """Check requesting vm id."""
        req_id = 'aed62b34-d1db-4153-b3d5-c2d23d267417'
        mocked_post.return_value = stubs.Response(json={'request_id': req_id})

        vdi_client = client.VdiClient(*self.default_credentials)
        self.assertEqual(vdi_client.request_vm(), req_id)

    @mock.patch('tionix_vdi_client.client.requests.post')
    def test_request_vm_no_request_id_in_response(self, mocked_post):
        """Check case when request id is not returned or empty."""
        vdi_client = client.VdiClient(*self.default_credentials)
        mocked_post.return_value = stubs.Response(json={'request_id': ''})

        self.assertRaises(exceptions.VDIServerResponseError,
                          vdi_client.request_vm)

    @mock.patch('tionix_vdi_client.client.requests.post')
    def test_get_vm_valid_response(self, mocked_post):
        """Requesting vm by id, valid response."""
        resp = {
            'server': 'test.stand', 'status': 'pending'
        }
        mocked_post.return_value = stubs.Response(json=resp)
        resp_expected = VM(resp['server'], resp['status'], None)

        vdi_client = client.VdiClient(*self.default_credentials)
        resp_got = vdi_client.get_vm('some-request-id')

        self.assertEqual(resp_expected, resp_got)

    @mock.patch('tionix_vdi_client.client.requests.post')
    def test_action_handle_server_errors(self, mocked_post):
        """Server-side errors are handled properly."""
        resp_codes_server_error = [400] + list(range(402, 431)) + list(range(500, 512))

        for err_code in resp_codes_server_error:
            resp = stubs.Response(json={'message': 'Unknown test err.'},
                                  status_code=err_code)
            mocked_post.return_value = resp
            for vdi_action in self.list_callable_vdi_actions():
                self.assertRaises(exceptions.VDIException, vdi_action)

    @mock.patch('tionix_vdi_client.client.requests.post')
    def test_actions_handle_authorisation_errors(self, mocked_post):
        """Authorisation error is handled."""
        resp_code_bad_credentials = 400
        resp = stubs.Response(
            json={'error': {'code': exceptions.ErrorCodes.AUTH_ERROR}},
            status_code=resp_code_bad_credentials,
        )
        mocked_post.return_value = resp

        for vdi_action in self.list_callable_vdi_actions():
            self.assertRaises(exceptions.VDIServerResponseError, vdi_action)

    @mock.patch('tionix_vdi_client.client.requests.post')
    def test_actions_handle_password_expired_error(self, mocked_post):
        """Password expire case raise proper error."""
        resp_code_bad_credentials = 400
        resp = stubs.Response(
            json={'error': {'code': exceptions.ErrorCodes.PASSWORD_EXPIRED}},
            status_code=resp_code_bad_credentials,
        )
        mocked_post.return_value = resp

        for vdi_action in self.list_callable_vdi_actions():
            self.assertRaises(exceptions.VDIServerResponseError, vdi_action)

    @mock.patch('tionix_vdi_client.client.requests.post')
    def test_action_handle_non_json_response(self, mocked_post):
        """Check behaviour for non-json responses."""
        resp = stubs.Response(content="<html>NotJson</html>")
        mocked_post.return_value = resp

        for vdi_action in self.list_callable_vdi_actions():
            self.assertRaises(exceptions.ApplicationConnectionError,
                              vdi_action)

    @mock.patch('tionix_vdi_client.client.requests.post')
    def test_actions_handle_not_found_unexpected_json_reply(self, mocked_post):
        """Check cases when server could return unexpected json response."""
        resp_code_not_found = 404
        resp = stubs.Response(json=["list", "of", "stuff"],
                              status_code=resp_code_not_found)
        mocked_post.return_value = resp

        for vdi_action in self.list_callable_vdi_actions():
            self.assertRaises(exceptions.VDIException, vdi_action)

    @mock.patch('tionix_vdi_client.client.requests.post')
    def test_actions_handle_connection_error(self, mocked_post):
        """Server connection errors are handled."""
        conn_err = requests.exceptions.ConnectionError()
        mocked_post.side_effect = conn_err
        for vdi_action in self.list_callable_vdi_actions():
            self.assertRaises(exceptions.ApplicationConnectionError,
                              vdi_action)

    @mock.patch('tionix_vdi_client.client.requests.post')
    def test_actions_handle_invalid_url(self, mocked_post):
        """Invalid urls in request are handled."""
        conn_err = requests.exceptions.InvalidURL()
        mocked_post.side_effect = conn_err
        for vdi_action in self.list_callable_vdi_actions():
            self.assertRaises(exceptions.ApplicationConnectionError,
                              vdi_action)

    @mock.patch('tionix_vdi_client.client.requests.post')
    def test_actions_when_first_cloud_not_available(self, mocked_post):
        """Second server should be used if the first one doesn't work."""
        self.conf.cloud = 'non-exists.loc'
        self.conf.secondary_cloud = 'existing.loc'
        response = stubs.Response(json={"request_id": 'some-id'})
        mocked_post.side_effect = [requests.exceptions.ConnectionError(),
                                   response]

        vdi_client = client.VdiClient(*self.default_credentials)
        vdi_client.request_vm()

        self.assertIn(self.conf.secondary_cloud, vdi_client.api_url)
        self.assertNotIn(self.conf.cloud, vdi_client.api_url)

    @mock.patch('tionix_vdi_client.client.requests.post')
    def test_request_vm_all_extra_params_passed_in_get(self, mocked_post):
        """Extra parameters are passed in get requests."""
        user, password = self.default_credentials[0:2]
        extra_params = {
            "domain_name": 'default',
            "project": 'test',
            "ip_address": '127.0.0.1',
        }
        expected_params = {
            'conn_type': CONNECTION_TYPE,
            'auth_tenant': 'test',
            'ip_address': '127.0.0.1',
            'project_domain_name': 'default',

        }
        response = stubs.Response(json={"request_id": 'some-id'})
        mocked_post.return_value = response

        vdi_client = client.VdiClient(user, password, **extra_params)
        vdi_client.request_vm()
        called_params_kwargs = mocked_post.call_args[1]
        self.assertDictContainsSubset(
            expected_params,
            called_params_kwargs['json'],
        )

    @mock.patch('tionix_vdi_client.client.requests.post')
    def test_request_vm_all_extra_params_passed_in_post(self, mocked_post):
        """Extra parameters are passed in get requests."""
        user, password = self.default_credentials[0:2]
        new_password = "some_password"
        extra_params = {
            "domain_name": 'default',
            "project": 'test',
            "ip_address": '127.0.0.1',
        }
        expected_params = {
            'new_password': new_password,
            'password': password,
        }
        response = stubs.Response(json={"request_id": 'some-id'})
        mocked_post.return_value = response

        vdi_client = client.VdiClient(user, password, **extra_params)
        vdi_client.change_password(new_password)
        called_params_kwargs = mocked_post.call_args[1]
        self.assertDictContainsSubset(
            expected_params,
            called_params_kwargs['json'],
        )


@unittest.skip('fix tests isolation, fails on review')
class TestVDIConnect(unittest.TestCase):
    def setUp(self):
        conf = stubs.TempConfig()
        conf.read()
        self.addCleanup(conf.delete_config)
        patcher = mock.patch('tionix_vdi_client.client.CONF', conf)
        self.conf = patcher.start()
        self.addCleanup(patcher.stop)

        self.vdi_client_mock = mock.MagicMock(spec=client.VdiClient)
        self.default_auth_params = {'user': 'tester', 'password': 'pass'}
        self.default_params = {'user_domain_name': 'default'}
        self.vdi_client_mock.auth_params = {}
        self.vdi_client_mock.auth_params.update(**self.default_auth_params)
        self.vdi_client_mock.auth_params.update(**self.default_params)

    @mock.patch('tionix_vdi_client.app_cli.rdp_connect')
    @mock.patch('tionix_vdi_client.app_cli.client.VdiClient')
    def test_vdi_connect_send_rdp_connect_command(self, vd_client_mock,
                                                  rdp_connect_mock):
        """RDP connection called properly for completed VM."""
        vm = {'ip': '127.0.0.1'}
        expected_call_params = self.default_auth_params.copy()
        expected_call_params.update({'vm_ip': '127.0.0.1',
                                     'domain_name': 'default',
                                     'rdp_token': 'token'})

        self.vdi_client_mock.get_vm.return_value = VM(
            vm, RequestStatus.COMPLETED, 'token')
        vd_client_mock.return_value = self.vdi_client_mock
        tionix_vdi_client.app_cli.main()

        rdp_connect_mock.assert_called_once_with(**expected_call_params)

    @mock.patch('tionix_vdi_client.app_cli.time.sleep')
    @mock.patch('tionix_vdi_client.app_cli.rdp_connect')
    @mock.patch('tionix_vdi_client.app_cli.client.VdiClient')
    def test_vdi_connect_vm_waits_till_available(self, vd_client_mock,
                                                 rdp_connect_mock,
                                                 sleep_mock):
        """If VM is pending we wait for it to become available."""
        vm_pending = VM({'ip': '127.0.0.1'}, RequestStatus.PENDING, '')
        vm_completed = VM({'ip': '127.0.0.1'}, RequestStatus.COMPLETED, '')
        self.vdi_client_mock.get_vm.side_effect = [vm_pending, vm_pending,
                                                   vm_completed]
        self.vdi_client_mock.rdp_token = 'some_token'
        vd_client_mock.return_value = self.vdi_client_mock
        tionix_vdi_client.app_cli.main()
        self.assertEqual(sleep_mock.call_count, 2)
        rdp_connect_mock.assert_called_once()
